@@ -1,7 +1,9 @@
 /*
  * %CopyrightBegin%
  *
- * Copyright Ericsson AB 2018-2024. All Rights Reserved.
+ * SPDX-License-Identifier: Apache-2.0
+ *
+ * Copyright Ericsson AB 2018-2025. All Rights Reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -167,6 +169,9 @@ typedef int BOOLEAN_T;
 
 #define B2S(__B__) ((__B__) ? "true" : "false")
 
+#define DOM2STR(__D__)   esock_domain_to_string((__D__))
+#define PROTO2STR(__P__) esock_protocol_to_string((__P__))
+
 #define TYPE2STR(T) (((T) == SOCK_STREAM) ? "stream" :                  \
                      (((T) == SOCK_DGRAM) ? "dgram"  :                  \
                       (((T) == SOCK_RAW)  ? "raw"    :                  \
@@ -177,6 +182,28 @@ typedef int BOOLEAN_T;
 #define NUM(Array) (sizeof(Array) / sizeof(*(Array)))
 
 
+/* ********************************************************************* *
+ *                              SOCKET and HANDLE                        *
+ * ********************************************************************* *
+ */
+
+#if defined(__WIN32__)
+
+#define INVALID_EVENT NULL
+#define SOCKET_FORMAT_STR "%lld"
+
+#else
+
+#define INVALID_HANDLE (-1)
+typedef int HANDLE;
+#define INVALID_SOCKET (-1)
+typedef int SOCKET; /* A subset of HANDLE */
+#define INVALID_EVENT INVALID_HANDLE
+#define SOCKET_FORMAT_STR "%d"
+
+#endif
+
+
 #ifdef HAVE_SOCKLEN_T
 #  define SOCKLEN_T socklen_t
 #else
@@ -205,6 +232,7 @@ typedef long ssize_t;
  */
 
 #define GLOBAL_ATOM_DEFS                       \
+    GLOBAL_ATOM_DEF(6to4);                     \
     GLOBAL_ATOM_DEF(abort);                    \
     GLOBAL_ATOM_DEF(accept);                   \
     GLOBAL_ATOM_DEF(acceptconn);               \
@@ -240,6 +268,7 @@ typedef long ssize_t;
     GLOBAL_ATOM_DEF(base_addr);                \
     GLOBAL_ATOM_DEF(bindtodevice);             \
     GLOBAL_ATOM_DEF(block_source);             \
+    GLOBAL_ATOM_DEF(bridge);                   \
     GLOBAL_ATOM_DEF(broadcast);                \
     GLOBAL_ATOM_DEF(bsp_state);                \
     GLOBAL_ATOM_DEF(busy_poll);                \
@@ -251,6 +280,7 @@ typedef long ssize_t;
     GLOBAL_ATOM_DEF(cancel);                   \
     GLOBAL_ATOM_DEF(cancelled);                \
     GLOBAL_ATOM_DEF(cantconfig);	       \
+    GLOBAL_ATOM_DEF(cellular);                 \
     GLOBAL_ATOM_DEF(chaos);                    \
     GLOBAL_ATOM_DEF(checksum);                 \
     GLOBAL_ATOM_DEF(close);                    \
@@ -282,6 +312,7 @@ typedef long ssize_t;
     GLOBAL_ATOM_DEF(delayed_ack_time);         \
     GLOBAL_ATOM_DEF(dgram);                    \
     GLOBAL_ATOM_DEF(dhcprunning);              \
+    GLOBAL_ATOM_DEF(disabled);                 \
     GLOBAL_ATOM_DEF(disable_fragments);        \
     GLOBAL_ATOM_DEF(dlci);                     \
     GLOBAL_ATOM_DEF(dma);                      \
@@ -300,6 +331,7 @@ typedef long ssize_t;
     GLOBAL_ATOM_DEF(eether);                   \
     GLOBAL_ATOM_DEF(efile);                    \
     GLOBAL_ATOM_DEF(egp);                      \
+    GLOBAL_ATOM_DEF(enabled);                  \
     GLOBAL_ATOM_DEF(enotsup);                  \
     GLOBAL_ATOM_DEF(eor);                      \
     GLOBAL_ATOM_DEF(error);                    \
@@ -326,7 +358,9 @@ typedef long ssize_t;
     GLOBAL_ATOM_DEF(frelay);                   \
     GLOBAL_ATOM_DEF(get_overlapped_result);    \
     GLOBAL_ATOM_DEF(get_peer_addr_info);       \
+    GLOBAL_ATOM_DEF(gif);                      \
     GLOBAL_ATOM_DEF(hatype);                   \
+    GLOBAL_ATOM_DEF(hdh1822);                  \
     GLOBAL_ATOM_DEF(hdrincl);                  \
     GLOBAL_ATOM_DEF(hmac_ident);               \
     GLOBAL_ATOM_DEF(hoplimit);                 \
@@ -353,6 +387,7 @@ typedef long ssize_t;
     GLOBAL_ATOM_DEF(ipcomp_level);             \
     GLOBAL_ATOM_DEF(ipip);                     \
     GLOBAL_ATOM_DEF(iplevel);                  \
+    GLOBAL_ATOM_DEF(ipv4);                     \
     GLOBAL_ATOM_DEF(ipv6);                     \
     GLOBAL_ATOM_DEF(irq);                      \
     GLOBAL_ATOM_DEF(i_want_mapped_v4_addr);    \
@@ -375,6 +410,7 @@ typedef long ssize_t;
     GLOBAL_ATOM_DEF(local);                    \
     GLOBAL_ATOM_DEF(localtlk);                 \
     GLOBAL_ATOM_DEF(local_auth_chunks);        \
+    GLOBAL_ATOM_DEF(loop);		       \
     GLOBAL_ATOM_DEF(loopback);		       \
     GLOBAL_ATOM_DEF(lowdelay);                 \
     GLOBAL_ATOM_DEF(lower_up);                 \
@@ -435,6 +471,7 @@ typedef long ssize_t;
     GLOBAL_ATOM_DEF(oobinline);                \
     GLOBAL_ATOM_DEF(options);                  \
     GLOBAL_ATOM_DEF(origdstaddr);              \
+    GLOBAL_ATOM_DEF(other);                    \
     GLOBAL_ATOM_DEF(otherhost);                \
     GLOBAL_ATOM_DEF(outgoing);                 \
     GLOBAL_ATOM_DEF(packet);                   \
@@ -449,11 +486,12 @@ typedef long ssize_t;
     GLOBAL_ATOM_DEF(pktinfo);                  \
     GLOBAL_ATOM_DEF(pktoptions);               \
     GLOBAL_ATOM_DEF(pkttype);                  \
-    GLOBAL_ATOM_DEF(ppromisc);		       \
     GLOBAL_ATOM_DEF(pointopoint);              \
     GLOBAL_ATOM_DEF(port);                     \
     GLOBAL_ATOM_DEF(portrange);                \
     GLOBAL_ATOM_DEF(portsel);                  \
+    GLOBAL_ATOM_DEF(ppromisc);		       \
+    GLOBAL_ATOM_DEF(ppp);                      \
     GLOBAL_ATOM_DEF(primary_addr);             \
     GLOBAL_ATOM_DEF(prim_file);                \
     GLOBAL_ATOM_DEF(priority);                 \
@@ -542,6 +580,7 @@ typedef long ssize_t;
     GLOBAL_ATOM_DEF(state);                    \
     GLOBAL_ATOM_DEF(status);                   \
     GLOBAL_ATOM_DEF(staticarp);		       \
+    GLOBAL_ATOM_DEF(stf);                      \
     GLOBAL_ATOM_DEF(stream);                   \
     GLOBAL_ATOM_DEF(syncnt);                   \
     GLOBAL_ATOM_DEF(syn_rcvd);                 \
@@ -585,6 +624,8 @@ typedef long ssize_t;
     GLOBAL_ATOM_DEF(write_pkg);                \
     GLOBAL_ATOM_DEF(write_tries);              \
     GLOBAL_ATOM_DEF(write_waits);              \
+    GLOBAL_ATOM_DEF(x25ddn);                   \
+    GLOBAL_ATOM_DEF(x25);                      \
     GLOBAL_ATOM_DEF(zero)
 
 
@@ -645,6 +686,7 @@ GLOBAL_ERROR_REASON_ATOM_DEFS;
 #define MKUL(E,UL)          enif_make_ulong((E), (UL))
 
 #define MCREATE(N)          enif_mutex_create((N))
+#define MCREATE3(P, B, S)   esock_mutex_create((P), (B), (S))
 #define MDESTROY(M)         enif_mutex_destroy((M))
 #define MLOCK(M)            enif_mutex_lock((M))
 #define MUNLOCK(M)          enif_mutex_unlock((M))
