@@ -1,6 +1,8 @@
 /*
  * %CopyrightBegin%
  *
+ * SPDX-License-Identifier: Apache-2.0
+ *
  * Copyright Ericsson AB 2018-2025. All Rights Reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
@@ -44,6 +46,10 @@
 #include <net/if_arp.h>
 #endif
 
+#if defined(HAVE_NET_IF_DL_H) && defined(AF_LINK)
+#include <net/if_types.h>
+#endif
+
 #include "socket_int.h"
 #include "sys.h"
 #include "socket_util.h"
@@ -146,6 +152,11 @@ static void make_sockaddr_dl(ErlNifEnv*    env,
 static SOCKLEN_T sa_local_length(int l, struct sockaddr_un* sa);
 #endif
 
+#if defined(HAVE_NET_IF_DL_H) && defined(AF_LINK)
+static ERL_NIF_TERM esock_encode_if_type(ErlNifEnv*   env,
+                                         unsigned int ifType);
+#endif
+
 
 /* The pre string has to be large enough (suggest 64) that
  * the name fits.
@@ -160,27 +171,16 @@ extern
 ErlNifMutex* esock_mutex_create(const char* pre, char* buf, SOCKET sock)
 {
 #if defined(ESOCK_VERBOSE_MTX_NAMES)
-    /* ESOCK_PRINTF("esock_mutex_create -> create name with" */
-    /*              "\r\n   pre:  %s" */
-    /*              "\r\n   sock: %d" */
-    /*              "\r\n", pre, sock); */
     sprintf(buf, "%s[" SOCKET_FORMAT_STR "]", pre, sock);
 #else
-    /* ESOCK_PRINTF("esock_mutex_create -> create name with" */
-    /*              "\r\n   pre: %s" */
-    /*              "\r\n", pre); */
     VOID(sock);
     sprintf(buf, "%s", pre);
 #endif
 
-    /* ESOCK_PRINTF("esock_mutex_create -> create mtx with name: %s" */
-    /*              "\r\n", buf); */
-
     return enif_mutex_create(buf);
 }
 
 
-
 /* *** esock_get_uint_from_map ***
  *
  * Simple utility function used to extract a unsigned int value from a map.
@@ -498,12 +498,18 @@ void esock_encode_sockaddr(ErlNifEnv*    env,
   switch (family) {
   case AF_INET:
       len = SALEN(addrLen, sizeof(struct sockaddr_in));
+      UDBG( ("SUTIL", "esock_encode_sockaddr -> 'inet' family addr: "
+             "\r\n   len: %d"
+             "\r\n", len) );
       esock_encode_sockaddr_in(env, &sockAddrP->in4, len, eSockAddr);
       break;
 
 #if defined(HAVE_IN6) && defined(AF_INET6)
   case AF_INET6:
       len = SALEN(addrLen, sizeof(struct sockaddr_in6));      
+      UDBG( ("SUTIL", "esock_encode_sockaddr -> 'inet6' family addr: "
+             "\r\n   len: %d"
+             "\r\n", len) );
       esock_encode_sockaddr_in6(env, &sockAddrP->in6, len, eSockAddr);
       break;
 #endif
@@ -511,6 +517,9 @@ void esock_encode_sockaddr(ErlNifEnv*    env,
 #ifdef HAS_AF_LOCAL
   case AF_LOCAL:
       len = sa_local_length(addrLen, &sockAddrP->un);
+      UDBG( ("SUTIL", "esock_encode_sockaddr -> 'local' family addr: "
+             "\r\n   len: %d"
+             "\r\n", len) );
       esock_encode_sockaddr_un(env, &sockAddrP->un, len, eSockAddr);
       break;
 #endif
@@ -518,6 +527,9 @@ void esock_encode_sockaddr(ErlNifEnv*    env,
 #ifdef AF_UNSPEC
   case AF_UNSPEC:
       len = SALEN(addrLen, 0);
+      UDBG( ("SUTIL", "esock_encode_sockaddr -> 'unspec' family addr: "
+             "\r\n   len: %d"
+             "\r\n", len) );
       esock_encode_sockaddr_native(env,
                                    &sockAddrP->sa, len,
                                    esock_atom_unspec,
@@ -528,6 +540,9 @@ void esock_encode_sockaddr(ErlNifEnv*    env,
 #if defined(HAVE_NETPACKET_PACKET_H)
   case AF_PACKET:
       len = SALEN(addrLen, sizeof(struct sockaddr_ll));      
+      UDBG( ("SUTIL", "esock_encode_sockaddr -> 'packet' family addr: "
+             "\r\n   len: %d"
+             "\r\n", len) );
       esock_encode_sockaddr_ll(env, &sockAddrP->ll, len, eSockAddr);
       break;
 #endif
@@ -535,6 +550,9 @@ void esock_encode_sockaddr(ErlNifEnv*    env,
 #if defined(AF_IMPLINK)
   case AF_IMPLINK:
       len = SALEN(addrLen, 0);
+      UDBG( ("SUTIL", "esock_encode_sockaddr -> 'implink' family addr: "
+             "\r\n   len: %d"
+             "\r\n", len) );
       esock_encode_sockaddr_native(env,
                                    &sockAddrP->sa, len,
                                    esock_atom_implink,
@@ -545,6 +563,9 @@ void esock_encode_sockaddr(ErlNifEnv*    env,
 #if defined(AF_PUP)
   case AF_PUP:
       len = SALEN(addrLen, 0);
+      UDBG( ("SUTIL", "esock_encode_sockaddr -> 'pup' family addr: "
+             "\r\n   len: %d"
+             "\r\n", len) );
       esock_encode_sockaddr_native(env,
                                    &sockAddrP->sa, len,
                                    esock_atom_pup,
@@ -555,6 +576,9 @@ void esock_encode_sockaddr(ErlNifEnv*    env,
 #if defined(AF_CHAOS)
   case AF_CHAOS:
       len = SALEN(addrLen, 0);
+      UDBG( ("SUTIL", "esock_encode_sockaddr -> 'chaos' family addr: "
+             "\r\n   len: %d"
+             "\r\n", len) );
       esock_encode_sockaddr_native(env,
                                    &sockAddrP->sa, len,
                                    esock_atom_chaos,
@@ -620,12 +644,18 @@ void esock_encode_sockaddr(ErlNifEnv*    env,
                   (CHARP(sockAddrP->dl.sdl_data) - CHARP(sockAddrP)) +
                   sockAddrP->dl.sdl_nlen + sockAddrP->dl.sdl_alen);
 #endif
+      UDBG( ("SUTIL", "esock_encode_sockaddr -> 'link' family addr: "
+             "\r\n   len: %d"
+             "\r\n", len) );
       esock_encode_sockaddr_dl(env, &sockAddrP->dl, len, eSockAddr);
     break;
 #endif
 
   default:
       len = SALEN(addrLen, 0);
+      UDBG( ("SUTIL", "esock_encode_sockaddr -> default (native) family addr: "
+             "\r\n   len: %d"
+             "\r\n", len) );
       esock_encode_sockaddr_native(env,
                                    &sockAddrP->sa, len,
                                    MKI(env, family),
@@ -658,77 +688,165 @@ void esock_encode_hwsockaddr(ErlNifEnv*       env,
 			     SOCKLEN_T        addrLen,
 			     ERL_NIF_TERM*    eSockAddr)
 {
-  ERL_NIF_TERM efamily;
-  int          family;
+    ERL_NIF_TERM efamily;
+    int          family;
 
-  // Sanity check
-  if (addrLen < (char *)&sockAddrP->sa_data - (char *)sockAddrP) {
-    // We got crap, cannot even know the address family
-    esock_encode_sockaddr_broken(env, sockAddrP, addrLen, eSockAddr);
-    return;
-  }
-  family = sockAddrP->sa_family;
+    // Sanity check
+    if (addrLen < (char *)&sockAddrP->sa_data - (char *)sockAddrP) {
+        // We got crap, cannot even know the address family
+        esock_encode_sockaddr_broken(env, sockAddrP, addrLen, eSockAddr);
+        return;
+    }
+    family = sockAddrP->sa_family;
 
-  UDBG( ("SUTIL", "esock_encode_hwsockaddr -> entry with"
-	 "\r\n   family:  %d"
-	 "\r\n   addrLen: %d"
-	 "\r\n", family, addrLen) );
+    UDBG( ("SUTIL", "esock_encode_hwsockaddr -> entry with"
+           "\r\n   family:  %d"
+           "\r\n   addrLen: %d"
+           "\r\n", family, addrLen) );
 
-  switch (family) {
+    switch (family) {
 #if defined(ARPHRD_NETROM)
-  case ARPHRD_NETROM:
-    efamily = esock_atom_netrom;
-    break;
+    case ARPHRD_NETROM:
+        efamily = esock_atom_netrom;
+        break;
 #endif
 
 #if defined(ARPHRD_ETHER)
-  case ARPHRD_ETHER:
-    efamily = esock_atom_ether;
-    break;
+    case ARPHRD_ETHER:
+        efamily = esock_atom_ether;
+        break;
 #endif
 
 #if defined(ARPHRD_IEEE802)
-  case ARPHRD_IEEE802:
-    efamily = esock_atom_ieee802;
-    break;
+    case ARPHRD_IEEE802:
+        efamily = esock_atom_ieee802;
+        break;
 #endif
 
 #if defined(ARPHRD_DLCI)
-  case ARPHRD_DLCI:
-    efamily = esock_atom_dlci;
-    break;
+    case ARPHRD_DLCI:
+        efamily = esock_atom_dlci;
+        break;
 #endif
 
 #if defined(ARPHRD_FRELAY)
-  case ARPHRD_FRELAY:
-    efamily = esock_atom_frelay;
-    break;
+    case ARPHRD_FRELAY:
+        efamily = esock_atom_frelay;
+        break;
 #endif
 
 #if defined(ARPHRD_IEEE1394)
-  case ARPHRD_IEEE1394:
-    efamily = esock_atom_ieee1394;
-    break;
+    case ARPHRD_IEEE1394:
+        efamily = esock_atom_ieee1394;
+        break;
 #endif
 
 #if defined(ARPHRD_LOOPBACK)
-  case ARPHRD_LOOPBACK:
-    efamily = esock_atom_loopback;
-    break;
+    case ARPHRD_LOOPBACK:
+        efamily = esock_atom_loopback;
+        break;
+#endif
+
+#if defined(ARPHRD_RAWIP)
+    case ARPHRD_RAWIP:
+        efamily = esock_atom_rawip;
+        break;
 #endif
 
 #if defined(ARPHRD_NONE)
-  case ARPHRD_NONE:
-    efamily = esock_atom_none;
-    break;
+    case ARPHRD_NONE:
+        efamily = esock_atom_none;
+        break;
 #endif
 
-  default:
-    efamily = MKI(env, family);
-    break;
-  }
+    default:
+        efamily = MKI(env, family);
+        break;
+    }
+
+    esock_encode_sockaddr_native(env, sockAddrP, addrLen, efamily, eSockAddr);
+}
+
+
+
+extern
+BOOLEAN_T esock_decode_hwsockaddr(ErlNifEnv*    env,
+                                  ERL_NIF_TERM  eSockAddr,
+                                  ESockAddress* sockAddrP,
+                                  SOCKLEN_T*    addrLen)
+{
+    ERL_NIF_TERM efamily;
+    int          family = -1;
+
+    if (!IS_MAP(env, eSockAddr))
+        return FALSE;
+
+    if (!GET_MAP_VAL(env, eSockAddr, esock_atom_family, &efamily))
+        return FALSE;
+
+    /* This is a bit ugly, but if-defing this properly is
+     * a bit messy so for now...
+     */
+#if defined(ARPHRD_NETROM)
+    if (IS_IDENTICAL(efamily, esock_atom_netrom)) {
+        family = ARPHRD_NETROM;
+    }
+#endif
+    
+#if defined(ARPHRD_ETHER)
+    if (IS_IDENTICAL(efamily, esock_atom_ether)) {
+        family = ARPHRD_ETHER;
+    }
+#endif
+
+#if defined(ARPHRD_IEEE802)
+    if (IS_IDENTICAL(efamily, esock_atom_ieee802)) {
+        family = ARPHRD_IEEE802;
+    }
+#endif
+
+#if defined(ARPHRD_DLCI)
+    if (IS_IDENTICAL(efamily, esock_atom_dlci)) {
+        family = ARPHRD_DLCI;
+    }
+#endif
+
+#if defined(ARPHRD_FRELAY)
+    if (IS_IDENTICAL(efamily, esock_atom_frelay)) {
+        family = ARPHRD_FRELAY;
+    }
+#endif
+
+#if defined(ARPHRD_IEEE1394)
+    if (IS_IDENTICAL(efamily, esock_atom_ieee1394)) {
+        family = ARPHRD_IEEE1394;
+    }
+#endif
+
+#if defined(ARPHRD_LOOPBACK)
+    if (IS_IDENTICAL(efamily, esock_atom_loopback)) {
+        family = ARPHRD_LOOPBACK;
+    }
+#endif
+
+#if defined(ARPHRD_RAWIP)
+    if (IS_IDENTICAL(efamily, esock_atom_rawip)) {
+        family = ARPHRD_RAWIP;
+    }
+#endif
+
+#if defined(ARPHRD_NONE)
+    if (IS_IDENTICAL(efamily, esock_atom_none)) {
+        family = ARPHRD_NONE;
+    }
+#endif
+
+    if (family == -1)
+	return FALSE;
+
+    return esock_decode_sockaddr_native(env, eSockAddr,
+                                        sockAddrP, family, addrLen);
 
-  esock_encode_sockaddr_native(env, sockAddrP, addrLen, efamily, eSockAddr);
 }
 
 
@@ -818,10 +936,16 @@ void esock_encode_sockaddr_in(ErlNifEnv*          env,
 {
     ERL_NIF_TERM ePort, eAddr;
     int          port;
+    /* The size of the actual data part, excluding padding */
+    SOCKLEN_T    minSz = sizeof(struct sockaddr_in) -
+        sizeof(sockAddrP->sin_zero);
 
-    UDBG( ("SUTIL", "esock_encode_sockaddr_in -> entry\r\n") );
+    UDBG( ("SUTIL", "esock_encode_sockaddr_in -> entry with"
+           "\r\n   addrLen:           %d"
+           "\r\n   required min size: %d"
+           "\r\n", addrLen, minSz) );
 
-    if (addrLen >= sizeof(struct sockaddr_in)) {
+    if (addrLen >= minSz) {
 
         /* The port */
         port  = ntohs(sockAddrP->sin_port);
@@ -837,7 +961,7 @@ void esock_encode_sockaddr_in(ErlNifEnv*          env,
         UDBG( ("SUTIL", "esock_encode_sockaddr_in -> wrong size: "
                "\r\n   addrLen:   %d"
                "\r\n   addr size: %d"
-               "\r\n", addrLen, sizeof(struct sockaddr_in)) );
+               "\r\n", addrLen, minSz) );
         esock_encode_sockaddr_native(env, (struct sockaddr *)sockAddrP,
                                      addrLen, esock_atom_inet, eSockAddr);
     }
@@ -1129,11 +1253,14 @@ void esock_encode_sockaddr_un(ErlNifEnv*          env,
  * (beside the mandatory family attribute, which is "inherited" from
  * the "sockaddr" type):
  *
- *    protocol: integer() (should be an atom really)
+ *    protocol: protocol()
  *    ifindex:  integer()
- *    hatype:   integer() (should be an atom really)
- *    pkttype:  integer() (should be an atom really)
- *    addr:     list()    (should be something useful...)
+ *    hatype:   hatype()
+ *    pkttype:  pkttype()
+ *    addr:     binary()
+ *
+ * Or if the address length is not enough, its encoded as a "native"
+ * packet (with type 'packet').
  *
  */
 
@@ -1210,32 +1337,46 @@ void esock_encode_sockaddr_dl(ErlNifEnv*          env,
 {
     ERL_NIF_TERM eindex, etype, enlen, ealen, eslen, edata;
     SOCKLEN_T    dlen;
+    SOCKLEN_T    ndsz = sizeof(struct sockaddr_dl)-sizeof(sockAddrP->sdl_data);
 
     UDBG( ("SUTIL", "esock_encode_sockaddr_dl -> entry with"
-           "\r\n.  addrLen: %d"
-           "\r\n", addrLen) );
+           "\r\n   addrLen:              %d"
+           "\r\n   non-data fields size: %d"
+           "\r\n", addrLen, ndsz) );
 
-    /* There is a minumum length (defined by the size of the data field) */
-    if (addrLen >= sizeof(struct sockaddr_dl)) {
+    /* Make sure the data field actually contains something */
+    if (addrLen >= ndsz) {
 
         /* index - if != 0, system given index for interface */
+        UDBG( ("SUTIL", "esock_encode_sockaddr_dl -> index: %d"
+               "\r\n", sockAddrP->sdl_index) );
         eindex = MKUI(env, sockAddrP->sdl_index);
 
         /* type -  interface type */
-        etype = MKUI(env, sockAddrP->sdl_type);
+        UDBG( ("SUTIL", "esock_encode_sockaddr_dl -> type: %d"
+               "\r\n", sockAddrP->sdl_type) );
+        etype = esock_encode_if_type(env, sockAddrP->sdl_type);
 
         /* nlen - interface name length, no trailing 0 reqd. */
+        UDBG( ("SUTIL", "esock_encode_sockaddr_dl -> nlen: %d"
+               "\r\n", sockAddrP->sdl_nlen) );
         enlen = MKUI(env, sockAddrP->sdl_nlen);
 
         /* alen - link level address length */
+        UDBG( ("SUTIL", "esock_encode_sockaddr_dl -> alen: %d"
+               "\r\n", sockAddrP->sdl_alen) );
         ealen = MKUI(env, sockAddrP->sdl_alen);
 
-        /* slen - ink layer selector length */
+        /* slen - link layer selector length */
+        UDBG( ("SUTIL", "esock_encode_sockaddr_dl -> slen: %d"
+               "\r\n", sockAddrP->sdl_slen) );
         eslen = MKUI(env, sockAddrP->sdl_slen);
 
         /* data - minimum work area, can be larger;    *
          *        contains both if name and ll address */
         dlen  = addrLen - (CHARP(sockAddrP->sdl_data) - CHARP(sockAddrP));
+        UDBG( ("SUTIL", "esock_encode_sockaddr_dl -> data len: %d"
+               "\r\n", dlen) );
         edata = esock_make_new_binary(env, &sockAddrP->sdl_data, dlen);
 
         make_sockaddr_dl(env,
@@ -1247,6 +1388,114 @@ void esock_encode_sockaddr_dl(ErlNifEnv*          env,
                                      addrLen, esock_atom_link, eSockAddr);
     }
 }
+
+
+static
+ERL_NIF_TERM esock_encode_if_type(ErlNifEnv*   env,
+                                  unsigned int ifType)
+{
+    ERL_NIF_TERM eIfType;
+
+    switch (ifType) {
+#if defined(IFT_OTHER)
+    case IFT_OTHER:
+        eIfType = esock_atom_other;
+        break;
+#endif
+
+#if defined(IFT_HDH1822)
+    case IFT_HDH1822:
+        eIfType = esock_atom_hdh1822;
+        break;
+#endif
+
+#if defined(IFT_X25DDN)
+    case IFT_X25DDN:
+        eIfType = esock_atom_x25ddn;
+        break;
+#endif
+
+#if defined(IFT_X25)
+    case IFT_X25:
+        eIfType = esock_atom_x25;
+        break;
+#endif
+
+#if defined(IFT_ETHER)
+    case IFT_ETHER:
+        eIfType = esock_atom_ether;
+        break;
+#endif
+
+#if defined(IFT_PPP)
+    case IFT_PPP:
+        eIfType = esock_atom_ppp;
+        break;
+#endif
+
+#if defined(IFT_LOOP)
+    case IFT_LOOP:
+        eIfType = esock_atom_loop;
+        break;
+#endif
+
+#if defined(IFT_IPV4)
+    case IFT_IPV4:
+        eIfType = esock_atom_ipv4;
+        break;
+#endif
+
+#if defined(IFT_IPV6)
+    case IFT_IPV6:
+        eIfType = esock_atom_ipv6;
+        break;
+#endif
+
+#if defined(IFT_6TO4)
+    case IFT_6TO4:
+        eIfType = esock_atom_6to4;
+        break;
+#endif
+
+#if defined(IFT_GIF)
+    case IFT_GIF:
+        eIfType = esock_atom_gif;
+        break;
+#endif
+
+#if defined(IFT_FAITH)
+    case IFT_FAITH:
+        eIfType = esock_atom_faith;
+        break;
+#endif
+
+#if defined(IFT_STF)
+    case IFT_STF:
+        eIfType = esock_atom_stf;
+        break;
+#endif
+
+#if defined(IFT_BRIDGE)
+    case IFT_BRIDGE:
+        eIfType = esock_atom_bridge;
+        break;
+#endif
+
+#if defined(IFT_CELLULAR)
+    case IFT_CELLULAR:
+        eIfType = esock_atom_cellular;
+        break;
+#endif
+
+    default:
+        eIfType = MKUI(env, ifType);
+        break;
+    }
+
+    return eIfType;
+}
+
+
 #endif
 
 
@@ -1647,6 +1896,100 @@ void esock_encode_domain(ErlNifEnv*    env,
 
 
 
+/*
+ * This is only intended for debugging.
+ * Transforms a 'domain' to a printable string.
+ */
+extern
+char* esock_domain_to_string(int domain)
+{
+    switch (domain) {
+    case AF_INET:
+        return "inet";
+        break;
+
+#if defined(HAVE_IN6) && defined(AF_INET6)
+    case AF_INET6:
+        return "inet6";
+        break;
+#endif
+
+#ifdef HAS_AF_LOCAL
+    case AF_LOCAL:
+        return "local";
+        break;
+#endif
+
+#ifdef AF_UNSPEC
+    case AF_UNSPEC:
+        return "unspec";
+        break;
+#endif
+
+    default:
+        return "undefined";
+    }
+}
+
+
+
+/*
+ * This is only intended for debugging.
+ * Transforms a 'protocol' to a printable string.
+ */
+extern
+char* esock_protocol_to_string(int protocol)
+{
+    switch (protocol) {
+#if defined(IPPROTO_IP)
+    case IPPROTO_IP:
+        return "ip";
+        break;
+#endif
+
+#if defined(IPPROTO_ICMP)
+    case IPPROTO_ICMP:
+        return "icmp";
+        break;
+#endif
+
+#if defined(IPPROTO_IGMP)
+    case IPPROTO_IGMP:
+        return "igmp";
+        break;
+#endif
+
+#if defined(IPPROTO_TCP)
+    case IPPROTO_TCP:
+        return "tcp";
+        break;
+#endif
+
+#if defined(IPPROTO_UDP)
+    case IPPROTO_UDP:
+        return "udp";
+        break;
+#endif
+
+#if defined(IPPROTO_SCTP)
+    case IPPROTO_SCTP:
+        return "sctp";
+        break;
+#endif
+
+#if defined(IPPROTO_RAW)
+    case IPPROTO_RAW:
+        return "raw";
+        break;
+#endif
+
+    default:
+        return "undefined";
+    }
+}
+
+
+
 /* +++ esock_decode_type +++
  *
  * Decode the Erlang form of the 'type' type, that is: 
@@ -2409,12 +2752,6 @@ ERL_NIF_TERM esock_errno_to_term(ErlNifEnv* env, int err)
         break;
 #endif
 
-#if defined(WSAENOTSOCK)
-    case WSAENOTSOCK:
-        return MKA(env, "enotsock");
-        break;
-#endif
-
 #if defined(ERROR_INVALID_NETNAME)
     case ERROR_INVALID_NETNAME:
         return MKA(env, "invalid_netname");
@@ -2589,24 +2926,7 @@ ERL_NIF_TERM esock_make_error_t2r(ErlNifEnv*   env,
                                   ERL_NIF_TERM tag,
                                   ERL_NIF_TERM reason)
 {
-    return esock_make_error(env, MKT2(env, tag, reason));
-}
-
-
-
-/* Create an 'error' two (2) tuple in the form:
- *
- *          {error, {invalid, {integer_range, I}}}
- *
- * The second arg (i) is already in the form of an
- * ERL_NIF_TERM so all we have to do is create the tuple.
- */
-extern
-ERL_NIF_TERM esock_make_error_integer_range(ErlNifEnv* env, ERL_NIF_TERM i)
-{
-    ERL_NIF_TERM intRange = MKT2(env, esock_atom_integer_range, i);
-
-    return esock_make_error_invalid(env, intRange);
+    return MKT2(env, esock_atom_error, MKT2(env, tag, reason));
 }
 
 
@@ -2618,7 +2938,25 @@ ERL_NIF_TERM esock_make_error_integer_range(ErlNifEnv* env, ERL_NIF_TERM i)
 extern
 ERL_NIF_TERM esock_make_error_invalid(ErlNifEnv* env, ERL_NIF_TERM what)
 {
-    return esock_make_error_t2r(env, esock_atom_invalid, what);
+    return MKT2(env,
+                esock_atom_error,
+                MKT2(env, esock_atom_invalid, what));
+}
+
+
+
+/* Create an 'error' two (2) tuple in the form:
+ *
+ *          {error, {invalid, {integer_range, I}}}
+ *
+ * The second element (i) is already in the form of an
+ * ERL_NIF_TERM so all we have to do is create the tuple.
+ */
+extern
+ERL_NIF_TERM esock_make_error_integer_range(ErlNifEnv* env, ERL_NIF_TERM i)
+{
+    return
+        esock_make_invalid(env, MKT2(env, esock_atom_integer_range, i));
 }
 
 
