@@ -1,7 +1,9 @@
 /*
  * %CopyrightBegin%
+ *
+ * SPDX-License-Identifier: Apache-2.0
  * 
- * Copyright Ericsson AB 2003-2016. All Rights Reserved.
+ * Copyright Ericsson AB 2003-2025. All Rights Reserved.
  * 
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -17,16 +19,13 @@
  * 
  * %CopyrightEnd%
  */
-#pragma comment(linker,"/manifestdependency:\"type='win32' "\
-		"name='Microsoft.Windows.Common-Controls' "\
-		"version='6.0.0.0' processorArchitecture='*' "\
-		"publicKeyToken='6595b64144ccf1df' language='*'\"")
+
 #include <windows.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include "init_file.h"
 
-typedef int ErlexecFunction(int, char **, HANDLE, int); 
+typedef int ErlexecFunction(int, char **, HANDLE);
 
 #define INI_FILENAME L"erl.ini"
 #define INI_SECTION "erlang"
@@ -38,18 +37,8 @@ static void error(char* format, ...);
 static wchar_t *erlexec_name;
 static wchar_t *erlexec_dir;
 
-#ifdef WIN32_WERL
-#define WERL 1
-int WINAPI wWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
-		    PWSTR szCmdLine, int iCmdShow)
-{
-    int argc = __argc;
-    wchar_t **argv = __wargv;
-#else
-#define WERL 0
 int wmain(int argc, wchar_t **argv)
 {
-#endif
   HANDLE erlexec_handle; /* Instance */
   ErlexecFunction *win_erlexec;
   wchar_t *path = malloc(100*sizeof(wchar_t));
@@ -77,7 +66,8 @@ int wmain(int argc, wchar_t **argv)
           wslpathlen = wcslen(wslpath);
       }
   }
-  pathlen = (wcslen(path) + wslpathlen + wcslen(erlexec_dir) + 2);
+  /* Add size for path delimiters and eos */
+  pathlen = (wcslen(path) + wslpathlen + wcslen(erlexec_dir) + 3);
   npath = (wchar_t *) malloc(pathlen*sizeof(wchar_t));
   if(wslpathlen > 0) {
       swprintf(npath,pathlen,L"%s;%s;%s",erlexec_dir,path,wslpath);
@@ -122,7 +112,7 @@ int wmain(int argc, wchar_t **argv)
 	}
 #endif
 
-  return (*win_erlexec)(argc,utf8argv,erlexec_handle,WERL);
+  return (*win_erlexec)(argc,utf8argv,erlexec_handle);
   
 } 
 
@@ -318,8 +308,8 @@ static void get_parameters(void)
     free(ini_filename);
 }
 
-
-static void error(char* format, ...)
+static __declspec(noreturn) void
+error(char* format, ...)
 {
     char sbuf[2048];
     va_list ap;
@@ -328,11 +318,6 @@ static void error(char* format, ...)
     vsprintf(sbuf, format, ap);
     va_end(ap);
 
-#ifndef WIN32_WERL 
-	fprintf(stderr, "%s\n", sbuf);
-#else
-	MessageBox(NULL, sbuf, "Werl", MB_OK|MB_ICONERROR);
-#endif
+    fprintf(stderr, "%s\n", sbuf);
     exit(1);
 }
-
