@@ -1,7 +1,9 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2008-2024. All Rights Reserved.
+%% SPDX-License-Identifier: Apache-2.0
+%%
+%% Copyright Ericsson AB 2008-2025. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -29,29 +31,21 @@
 This module provides API functions to send SSH Connection Protocol events to the
 other side of an SSH channel.
 
-The [SSH Connection Protocol](http://www.ietf.org/rfc/rfc4254.txt) is used by
+The [SSH Connection Protocol (RFC 4254)](http://www.ietf.org/rfc/rfc4254.txt) is used by
 clients and servers, that is, SSH channels, to communicate over the SSH
 connection. The API functions in this module send SSH Connection Protocol
 events, which are received as messages by the remote channel handling the remote
 channel. The Erlang format of thoose messages is (see also
 [below](`t:event/0`)):
 
-`{ssh_cm, ``t:ssh:connection_ref/0``, ``t:channel_msg/0``}`
+`{ssh_cm,` `t:ssh:connection_ref/0` `,` `t:channel_msg/0` `}`
 
 If the `m:ssh_client_channel` behavior is used to implement the channel process,
 these messages are handled by
 [handle_ssh_msg/2](`c:ssh_client_channel:handle_ssh_msg/2`).
 """.
--moduledoc(#{titles =>
-                 [{type,<<"SSH Connection Protocol: General">>},
-                  {type,<<"Data Transfer (RFC 4254, section 5.2)">>},
-                  {type,<<"Closing a Channel (RFC 4254, section 5.3)">>},
-                  {type,<<"Requesting a Pseudo-Terminal (RFC 4254, section 6.2)">>},
-                  {type,<<"Environment Variable Passing (RFC 4254, section 6.4)">>},
-                  {type,<<"Starting a Shell or Command (RFC 4254, section 6.5)">>},
-                  {type,<<"Window Dimension Change Message (RFC 4254, section 6.7)">>},
-                  {type,<<"Signals (RFC 4254, section 6.9)">>},
-                  {type,<<"Returning Exit Status (RFC 4254, section 6.10)">>}]}).
+
+-include_lib("kernel/include/logger.hrl").
 
 -include("ssh.hrl").
 -include("ssh_connect.hrl").
@@ -122,12 +116,11 @@ If not, the `t:reason/0` indicates what went wrong:
 """.
 -type reason() :: closed | timeout .
 
--doc(#{equiv => reason/0}).
 -type result() :: req_status() | {error, reason()} .
 
 -doc """
 The valid values are `0` ("normal") and `1` ("stderr"), see
-[RFC 4254, Section 5.2](https://tools.ietf.org/html/rfc4254#page/8).
+[RFC 4254, Section 5.2](https://tools.ietf.org/html/rfc4254#page-8).
 """.
 -type ssh_data_type_code() :: non_neg_integer(). % Only 0 and 1 are used
 
@@ -150,8 +143,7 @@ The valid values are `0` ("normal") and `1` ("stderr"), see
               exec_ch_msg/0
              ]).
 
--doc(#{title => <<"SSH Connection Protocol: General">>,
-       equiv => channel_msg/0}).
+-doc(#{group => <<"General">>}).
 -type event() :: {ssh_cm, ssh:connection_ref(), channel_msg()}.
 -doc """
 As mentioned in the introduction, the
@@ -160,7 +152,7 @@ handled as messages. When writing a channel handling process without using the
 support by the `m:ssh_client_channel` behavior the process must handle thoose
 messages.
 """.
--doc(#{title => <<"SSH Connection Protocol: General">>}).
+-doc(#{group => <<"General">>}).
 -type channel_msg() ::  data_ch_msg()
                       | eof_ch_msg()
                       | closed_ch_msg()
@@ -179,14 +171,15 @@ Messages that include a `WantReply` expect the channel handling process to call
 [ssh_connection:reply_request/4](`reply_request/4`) with the boolean value of
 `WantReply` as the second argument.
 """.
--doc(#{title => <<"SSH Connection Protocol: General">>}).
+-doc(#{group => <<"General">>}).
 -type want_reply() :: boolean().
 
 -doc """
 Data has arrived on the channel. This event is sent as a result of calling
 [ssh_connection:send/3,4,5](`send/3`).
 """.
--doc(#{title => <<"Data Transfer (RFC 4254, section 5.2)">>}).
+-doc(#{group => <<"Data Transfer">>,
+       rfc => ~"RFC 4254, section 5.2"}).
 -type data_ch_msg() :: {data,
                         ssh:channel_id(),
                         ssh_data_type_code(),
@@ -196,7 +189,8 @@ Data has arrived on the channel. This event is sent as a result of calling
 Indicates that the other side sends no more data. This event is sent as a result
 of calling [ssh_connection:send_eof/2](`send_eof/2`).
 """.
--doc(#{title => <<"Closing a Channel (RFC 4254, section 5.3)">>}).
+-doc(#{group => <<"Closing a Channel">>,
+       rfc => ~"RFC 4254, section 5.3"}).
 -type eof_ch_msg() :: {eof,
                        ssh:channel_id()
                       } .
@@ -207,7 +201,8 @@ this message. There is currently no function to generate this event as the
 signals referred to are on OS-level and not something generated by an Erlang
 program.
 """.
--doc(#{title => <<"Signals (RFC 4254, section 6.9)">>}).
+-doc(#{group => <<"Signals">>,
+       rfc => ~"RFC 4254, section 6.9"}).
 -type signal_ch_msg() :: {signal,
                           ssh:channel_id(),
                           SignalName :: string()
@@ -218,7 +213,8 @@ message can be received. For details on valid string values, see
 [RFC 4254](https://tools.ietf.org/html/rfc4254#section-6.10) Section 6.10, which
 shows a special case of these signals.
 """.
--doc(#{title => <<"Returning Exit Status (RFC 4254, section 6.10)">>}).
+-doc(#{group => <<"Exit Status">>,
+       rfc => ~"RFC 4254, section 6.10"}).
 -type exit_signal_ch_msg() :: {exit_signal, ssh:channel_id(),
                                ExitSignal :: string(),
                                ErrorMsg :: string(),
@@ -229,7 +225,8 @@ be sent to return the exit status of the command. A zero `exit_status` usually
 means that the command terminated successfully. This event is sent as a result
 of calling [ssh_connection:exit_status/3](`exit_status/3`).
 """.
--doc(#{title => <<"Returning Exit Status (RFC 4254, section 6.10)">>}).
+-doc(#{group => <<"Exit Status">>,
+       rfc => ~"RFC 4254, section 6.10"}).
 -type exit_status_ch_msg() :: {exit_status,
                                ssh:channel_id(),
                                ExitStatus :: non_neg_integer()
@@ -239,7 +236,8 @@ This event is sent as a result of calling [ssh_connection:close/2](`close/2`).
 Both the handling of this event and sending it are taken care of by the
 `m:ssh_client_channel` behavior.
 """.
--doc(#{title => <<"Closing a Channel (RFC 4254, section 5.3)">>}).
+-doc(#{group => <<"Closing a Channel">>,
+       rfc => ~"RFC 4254, section 5.3"}).
 -type closed_ch_msg() :: {closed,
                           ssh:channel_id()
                          } .
@@ -247,15 +245,16 @@ Both the handling of this event and sending it are taken care of by the
 Environment variables can be passed to the shell/command to be started later.
 This event is sent as a result of calling [ssh_connection:setenv/5](`setenv/5`).
 """.
--doc(#{title => <<"Environment Variable Passing (RFC 4254, section 6.4)">>}).
+-doc(#{group => <<"Environment Variable">>,
+       rfc => ~"RFC 4254, section 6.4"}).
 -type env_ch_msg() :: {env,
                        ssh:channel_id(),
                        want_reply(),
                        Var :: string(),
                        Value :: string()
                       } .
--doc(#{title => <<"Requesting a Pseudo-Terminal (RFC 4254, section 6.2)">>,
-       equiv => term_mode/0}).
+-doc(#{group => <<"Pseudo-Terminal">>,
+       rfc => ~"RFC 4254, section 6.2"}).
 -type pty_ch_msg() :: {pty,
                        ssh:channel_id(),
                        want_reply(),
@@ -280,7 +279,8 @@ also be an `Opcode` if the mnemonic name is not listed in the RFC. Example:
 `OP code: 53, mnemonic name ECHO erlang atom: echo`. This event is sent as a
 result of calling [ssh_connection:ptty_alloc/4](`ptty_alloc/4`).
 """.
--doc(#{title => <<"Requesting a Pseudo-Terminal (RFC 4254, section 6.2)">>}).
+-doc(#{group => <<"Pseudo-Terminal">>,
+       rfc => ~"RFC 4254, section 6.2"}).
 -type term_mode() :: {Opcode :: atom() | byte(),
                       Value :: non_neg_integer()} .
 
@@ -288,7 +288,8 @@ result of calling [ssh_connection:ptty_alloc/4](`ptty_alloc/4`).
 This message requests that the user default shell is started at the other end.
 This event is sent as a result of calling [ssh_connection:shell/2](`shell/2`).
 """.
--doc(#{title => <<"Starting a Shell or Command (RFC 4254, section 6.5)">>}).
+-doc(#{group => <<"Shell or Command">>,
+       rfc => ~"RFC 4254, section 6.5"}).
 -type shell_ch_msg() :: {shell,
                          ssh:channel_id(),
                          want_reply()
@@ -298,7 +299,8 @@ When the window (terminal) size changes on the client side, it _can_ send a
 message to the server side to inform it of the new dimensions. No API function
 generates this event.
 """.
--doc(#{title => <<"Window Dimension Change Message (RFC 4254, section 6.7)">>}).
+-doc(#{group => <<"Window Change">>,
+       rfc => ~"RFC 4254, section 6.7"}).
 -type window_change_ch_msg() :: {window_change,
                                  ssh:channel_id(),
                                  CharWidth :: non_neg_integer(),
@@ -310,7 +312,8 @@ generates this event.
 This message requests that the server starts execution of the given command.
 This event is sent as a result of calling [ssh_connection:exec/4 ](`exec/4`).
 """.
--doc(#{title => <<"Starting a Shell or Command (RFC 4254, section 6.5)">>}).
+-doc(#{group => <<"Shell or Command">>,
+       rfc => ~"RFC 4254, section 6.5"}).
 -type exec_ch_msg() :: {exec,
                         ssh:channel_id(),
                         want_reply(),
@@ -340,8 +343,8 @@ dummy(_) -> false.
       Timeout :: timeout(),
       Result :: {ok, ssh:channel_id()} | {error, reason()} .
 
-session_channel(ConnectionHandler, Timeout) ->
-    session_channel(ConnectionHandler, undefined, undefined, Timeout).
+session_channel(ConnectionRef, Timeout) ->
+    session_channel(ConnectionRef, undefined, undefined, Timeout).
 
 
 -doc """
@@ -355,8 +358,8 @@ is the id used as input to the other functions in this module.
       Timeout :: timeout(),
       Result :: {ok, ssh:channel_id()} | {error, reason()} .
 
-session_channel(ConnectionHandler, InitialWindowSize, MaxPacketSize, Timeout) ->
-    open_channel(ConnectionHandler, "session", <<>>,
+session_channel(ConnectionRef, InitialWindowSize, MaxPacketSize, Timeout) ->
+    open_channel(ConnectionRef, "session", <<>>,
                  InitialWindowSize,
                  MaxPacketSize,
                  Timeout).
@@ -365,11 +368,11 @@ session_channel(ConnectionHandler, InitialWindowSize, MaxPacketSize, Timeout) ->
 %% Description: Opens a channel for the given type.
 %% --------------------------------------------------------------------
 -doc false.
-open_channel(ConnectionHandler, Type, ChanData, Timeout) ->
-    open_channel(ConnectionHandler, Type, ChanData, undefined, undefined, Timeout).
+open_channel(ConnectionRef, Type, ChanData, Timeout) ->
+    open_channel(ConnectionRef, Type, ChanData, undefined, undefined, Timeout).
 
-open_channel(ConnectionHandler, Type, ChanData, InitialWindowSize, MaxPacketSize, Timeout) ->
-    case ssh_connection_handler:open_channel(ConnectionHandler, Type, ChanData,
+open_channel(ConnectionRef, Type, ChanData, InitialWindowSize, MaxPacketSize, Timeout) ->
+    case ssh_connection_handler:open_channel(ConnectionRef, Type, ChanData,
                                              InitialWindowSize, MaxPacketSize,
                                              Timeout) of
         {open, Channel} ->
@@ -407,6 +410,12 @@ request is a one-time execution that closes the channel when it is done.
 
 See the User's Guide section on
 [One-Time Execution](using_ssh.md#one-time-execution) for examples.
+
+> #### Note {: .info }
+>
+> In case when command generates large amount of output data, manual
+> window adjustment might be necessary in order to receive it.
+> see [`ssh_connectino:adjust_window/3`](`adjust_window/3`)
 """.
 -spec exec(ConnectionRef, ChannelId, Command, Timeout) -> result() when
       ConnectionRef :: ssh:connection_ref(),
@@ -414,8 +423,8 @@ See the User's Guide section on
       Command :: string(),
       Timeout :: timeout().
 
-exec(ConnectionHandler, ChannelId, Command, TimeOut) ->
-    ssh_connection_handler:request(ConnectionHandler, self(), ChannelId, "exec",
+exec(ConnectionRef, ChannelId, Command, TimeOut) ->
+    ssh_connection_handler:request(ConnectionRef, self(), ChannelId, "exec",
 				   true, [?string(Command)], TimeOut).
 
 %%--------------------------------------------------------------------
@@ -437,8 +446,8 @@ would break a large number of existing software.
       ChannelId :: ssh:channel_id(),
       Result :: ok | success | failure | {error, timeout} .
 
-shell(ConnectionHandler, ChannelId) ->
-    ssh_connection_handler:request(ConnectionHandler, self(), ChannelId,
+shell(ConnectionRef, ChannelId) ->
+    ssh_connection_handler:request(ConnectionRef, self(), ChannelId,
  				   "shell", false, <<>>, 0).
 %%--------------------------------------------------------------------
 %%
@@ -457,8 +466,8 @@ The function [`subsystem/4`](`subsystem/4`) and subsequent calls of
       Subsystem  :: string(),
       Timeout :: timeout().
 
-subsystem(ConnectionHandler, ChannelId, SubSystem, TimeOut) ->
-     ssh_connection_handler:request(ConnectionHandler, self(),
+subsystem(ConnectionRef, ChannelId, SubSystem, TimeOut) ->
+     ssh_connection_handler:request(ConnectionRef, self(),
 				    ChannelId, "subsystem", 
 				    true, [?string(SubSystem)], TimeOut).
 %%--------------------------------------------------------------------
@@ -466,51 +475,51 @@ subsystem(ConnectionHandler, ChannelId, SubSystem, TimeOut) ->
 %%--------------------------------------------------------------------
 -doc(#{equiv => send/5}).
 -spec send(connection_ref(), channel_id(), iodata()) ->
-		  ok | {error, timeout | closed}.
+		  ok | {error, reason()}.
 
-send(ConnectionHandler, ChannelId, Data) ->
-    send(ConnectionHandler, ChannelId, 0, Data, infinity).
+send(ConnectionRef, ChannelId, Data) ->
+    send(ConnectionRef, ChannelId, 0, Data, infinity).
 
 
 -doc """
 send(ConnectionRef, ChannelId, Type, Data)
 
+Depending on input arguments equivalent to one of `send/5` calls specified below.
+
 Equivalent to [send(ConnectionRef, ChannelId, 0, Data, TimeOut)](`send/5`) if
 called with TimeOut being integer.
 
 Equivalent to [send(ConnectionRef, ChannelId, 0, Data, infinity)](`send/5`) if
 called with TimeOut being infinity atom.
 
-Equivalent to [send(ConnectionHandler, ChannelId, Type, Data, infinity)](`send/5`) if
+Equivalent to [send(ConnectionRef, ChannelId, Type, Data, infinity)](`send/5`) if
 called with last argument which is not integer or infinity atom.
 """.
 
--spec send(connection_ref(), channel_id(), iodata(), timeout()) -> ok |  {error, reason()};
-          (connection_ref(), channel_id(), ssh_data_type_code(), iodata()) -> ok |  {error, reason()}.
+-spec send(connection_ref(), channel_id(), iodata(), timeout()) -> ok | {error, reason()};
+          (connection_ref(), channel_id(), ssh_data_type_code(), iodata()) -> ok | {error, reason()}.
 
-send(ConnectionHandler, ChannelId, Data, TimeOut) when is_integer(TimeOut) ->
-    send(ConnectionHandler, ChannelId, 0, Data, TimeOut);
+send(ConnectionRef, ChannelId, Data, TimeOut) when is_integer(TimeOut) ->
+    send(ConnectionRef, ChannelId, 0, Data, TimeOut);
 
-send(ConnectionHandler, ChannelId, Data, infinity) ->
-    send(ConnectionHandler, ChannelId, 0, Data, infinity);
+send(ConnectionRef, ChannelId, Data, infinity) ->
+    send(ConnectionRef, ChannelId, 0, Data, infinity);
 
-send(ConnectionHandler, ChannelId, Type, Data) ->
-    send(ConnectionHandler, ChannelId, Type, Data, infinity).
+send(ConnectionRef, ChannelId, Type, Data) ->
+    send(ConnectionRef, ChannelId, Type, Data, infinity).
 
 
 -doc """
-send(ConnectionRef, ChannelId, Type, Data, TimeOut) -> ok | Error
-
 Is to be called by client- and server-channel processes to send data to each
 other.
 
 The function `subsystem/4` and subsequent calls of `send/3,4,5` must be executed
 in the same process.
 """.
--spec send(connection_ref(), channel_id(), ssh_data_type_code(), iodata(), timeout()) -> ok |  {error, reason()}.
+-spec send(connection_ref(), channel_id(), ssh_data_type_code(), iodata(), timeout()) -> ok | {error, reason()}.
 
-send(ConnectionHandler, ChannelId, Type, Data, TimeOut) ->
-    ssh_connection_handler:send(ConnectionHandler, ChannelId,
+send(ConnectionRef, ChannelId, Type, Data, TimeOut) ->
+    ssh_connection_handler:send(ConnectionRef, ChannelId,
 				Type, Data, TimeOut).
 %%--------------------------------------------------------------------
 -doc "Sends EOF on channel `ChannelId`.".
@@ -521,8 +530,8 @@ send(ConnectionHandler, ChannelId, Type, Data, TimeOut) ->
 %%
 %% Description: Sends eof on the channel <ChannelId>.
 %%--------------------------------------------------------------------
-send_eof(ConnectionHandler, Channel) ->
-    ssh_connection_handler:send_eof(ConnectionHandler, Channel).
+send_eof(ConnectionRef, Channel) ->
+    ssh_connection_handler:send_eof(ConnectionRef, Channel).
 
 %%--------------------------------------------------------------------
 -doc """
@@ -545,8 +554,8 @@ server-side channel processes.
 %%
 %% Description: Adjusts the ssh flowcontrol window.
 %%--------------------------------------------------------------------
-adjust_window(ConnectionHandler, Channel, Bytes) ->
-    ssh_connection_handler:adjust_window(ConnectionHandler, Channel, Bytes).
+adjust_window(ConnectionRef, Channel, Bytes) ->
+    ssh_connection_handler:adjust_window(ConnectionRef, Channel, Bytes).
 
 %%--------------------------------------------------------------------
 -doc """
@@ -563,11 +572,11 @@ called by a client channel processes.
 %%
 %% Description: Environment variables may be passed to the shell/command to be
 %% started later.
-setenv(ConnectionHandler, ChannelId, Var, Value, TimeOut) ->
-    setenv(ConnectionHandler, ChannelId, true, Var, Value, TimeOut).
+setenv(ConnectionRef, ChannelId, Var, Value, TimeOut) ->
+    setenv(ConnectionRef, ChannelId, true, Var, Value, TimeOut).
 
-setenv(ConnectionHandler, ChannelId, WantReply, Var, Value, TimeOut) ->
-    case ssh_connection_handler:request(ConnectionHandler, ChannelId,
+setenv(ConnectionRef, ChannelId, WantReply, Var, Value, TimeOut) ->
+    case ssh_connection_handler:request(ConnectionRef, ChannelId,
                                         "env", WantReply,
                                         [?string(Var), ?string(Value)], TimeOut) of
         ok when WantReply == false ->
@@ -593,8 +602,8 @@ a close event.
 %%
 %% Description: Sends a close message on the channel <ChannelId>.
 %%--------------------------------------------------------------------
-close(ConnectionHandler, ChannelId) ->
-    ssh_connection_handler:close(ConnectionHandler, ChannelId).
+close(ConnectionRef, ChannelId) ->
+    ssh_connection_handler:close(ConnectionRef, ChannelId).
 
 %%--------------------------------------------------------------------
 -doc """
@@ -612,8 +621,8 @@ Protocol message containing a `WantReply` boolean value.
 %%
 %% Description: Send status replies to requests that want such replies.
 %%--------------------------------------------------------------------
-reply_request(ConnectionHandler, true, Status, ChannelId) ->
-    ssh_connection_handler:reply_request(ConnectionHandler, Status, ChannelId);
+reply_request(ConnectionRef, true, Status, ChannelId) ->
+    ssh_connection_handler:reply_request(ConnectionRef, Status, ChannelId);
 reply_request(_,false, _, _) ->
     ok.
 
@@ -627,8 +636,8 @@ reply_request(_,false, _, _) ->
       ChannelId :: ssh:channel_id(),
       Options  :: proplists:proplist().
 
-ptty_alloc(ConnectionHandler, Channel, Options) ->
-    ptty_alloc(ConnectionHandler, Channel, Options, infinity).
+ptty_alloc(ConnectionRef, Channel, Options) ->
+    ptty_alloc(ConnectionRef, Channel, Options, infinity).
 
 
 -doc """
@@ -659,11 +668,11 @@ Options:
       Options  :: proplists:proplist(),
       Timeout :: timeout().
 
-ptty_alloc(ConnectionHandler, Channel, Options0, TimeOut) ->
+ptty_alloc(ConnectionRef, Channel, Options0, TimeOut) ->
     TermData = backwards_compatible(Options0, []), % FIXME
     {Width, PixWidth} = pty_default_dimensions(width, TermData),
     {Height, PixHeight} = pty_default_dimensions(height, TermData),
-    pty_req(ConnectionHandler, Channel,
+    pty_req(ConnectionRef, Channel,
 	    proplists:get_value(term, TermData, os:getenv("TERM", ?DEFAULT_TERMINAL)),
 	    proplists:get_value(width, TermData, Width),
 	    proplists:get_value(height, TermData, Height),
@@ -678,19 +687,19 @@ ptty_alloc(ConnectionHandler, Channel, Options0, TimeOut) ->
 %% Should they be documented and tested?
 %%--------------------------------------------------------------------
 -doc false.
-window_change(ConnectionHandler, Channel, Width, Height) ->
-    window_change(ConnectionHandler, Channel, Width, Height, 0, 0).
+window_change(ConnectionRef, Channel, Width, Height) ->
+    window_change(ConnectionRef, Channel, Width, Height, 0, 0).
 -doc false.
-window_change(ConnectionHandler, Channel, Width, Height,
+window_change(ConnectionRef, Channel, Width, Height,
 	      PixWidth, PixHeight) ->
-    ssh_connection_handler:request(ConnectionHandler, Channel,
+    ssh_connection_handler:request(ConnectionRef, Channel,
 				   "window-change", false, 
 				   [?uint32(Width), ?uint32(Height),
 				    ?uint32(PixWidth), ?uint32(PixHeight)], 0).
 
 -doc false.
-signal(ConnectionHandler, Channel, Sig) ->
-    ssh_connection_handler:request(ConnectionHandler, Channel,
+signal(ConnectionRef, Channel, Sig) ->
+    ssh_connection_handler:request(ConnectionRef, Channel,
 				   "signal", false, [?string(Sig)], 0).
 
 
@@ -702,8 +711,8 @@ to the client.
       ConnectionRef :: ssh:connection_ref(),
       ChannelId :: ssh:channel_id(),
       Status  :: integer().
-exit_status(ConnectionHandler, Channel, Status) ->
-    ssh_connection_handler:request(ConnectionHandler, Channel,
+exit_status(ConnectionRef, Channel, Status) ->
+    ssh_connection_handler:request(ConnectionRef, Channel,
 				   "exit-status", false, [?uint32(Status)], 0).
 
 %%--------------------------------------------------------------------
@@ -746,6 +755,25 @@ channel_data(ChannelId, DataType, Data0,
 %%%
 
 -doc false.
+handle_msg(#ssh_msg_disconnect{code = Code, description = Description}, Connection, _, _SSH) ->
+    {disconnect, {Code, Description}, handle_stop(Connection)};
+
+handle_msg(Msg, Connection, server, Ssh = #ssh{authenticated = false}) ->
+    %% See RFC4252 6.
+    %% Message numbers of 80 and higher are reserved for protocols running
+    %% after this authentication protocol, so receiving one of them before
+    %% authentication is complete is an error, to which the server MUST
+    %% respond by disconnecting, preferably with a proper disconnect message
+    %% sent to ease troubleshooting.
+    MsgFun = fun(M) ->
+                     MaxLogItemLen = ?GET_OPT(max_log_item_len, Ssh#ssh.opts),
+                     io_lib:format("Connection terminated. Unexpected message for unauthenticated user."
+                                   " Message:  ~w", [M],
+                                   [{chars_limit, MaxLogItemLen}])
+             end,
+    ?LOG_DEBUG(MsgFun, [Msg]),
+    {disconnect, {?SSH_DISCONNECT_PROTOCOL_ERROR, "Connection refused"}, handle_stop(Connection)};
+
 handle_msg(#ssh_msg_channel_open_confirmation{recipient_channel = ChannelId, 
 					      sender_channel = RemoteId,
 					      initial_window_size = WindowSz,
@@ -826,21 +854,23 @@ handle_msg(#ssh_msg_channel_extended_data{recipient_channel = ChannelId,
     channel_data_reply_msg(ChannelId, Connection, DataType, Data);
 
 handle_msg(#ssh_msg_channel_window_adjust{recipient_channel = ChannelId,
-					  bytes_to_add = Add}, 
+					  bytes_to_add = Add},
 	   #connection{channel_cache = Cache} = Connection, _, _SSH) ->
-    #channel{send_window_size = Size, remote_id = RemoteId} = 
-	Channel0 = ssh_client_channel:cache_lookup(Cache, ChannelId), 
-    
-    {SendList, Channel} =  %% TODO: Datatype 0 ?
-	update_send_window(Channel0#channel{send_window_size = Size + Add},
-			   0, undefined, Connection),
-    
-    Replies = lists:map(fun({Type, Data}) -> 
-				{connection_reply, channel_data_msg(RemoteId, Type, Data)}
-			end, SendList),
-    FlowCtrlMsgs = flow_control(Channel, Cache),
-    {Replies ++ FlowCtrlMsgs, Connection};
-
+    case ssh_client_channel:cache_lookup(Cache, ChannelId) of
+        Channel0 = #channel{send_window_size = Size,
+                            remote_id = RemoteId} ->
+            {SendList, Channel} =  %% TODO: Datatype 0 ?
+                update_send_window(Channel0#channel{send_window_size = Size + Add},
+                                   0, undefined, Connection),
+            Replies = lists:map(fun({Type, Data}) ->
+                                        {connection_reply,
+                                         channel_data_msg(RemoteId, Type, Data)}
+                                end, SendList),
+            FlowCtrlMsgs = flow_control(Channel, Cache),
+            {Replies ++ FlowCtrlMsgs, Connection};
+        undefined ->
+            {[], Connection}
+    end;
 handle_msg(#ssh_msg_channel_open{channel_type = "session" = Type,
 				 sender_channel = RemoteId,
 				 initial_window_size = WindowSz,
@@ -884,7 +914,7 @@ handle_msg(#ssh_msg_channel_open{channel_type = "forwarded-tcpip",
                        suggest_window_size = WinSz,
                        suggest_packet_size = PktSz,
                        options = Options,
-                       sub_system_supervisor = SubSysSup
+                       connection_supervisor = ConnectionSup
                       } = C,
 	   client, _SSH) ->
     {ReplyMsg, NextChId} =
@@ -892,7 +922,7 @@ handle_msg(#ssh_msg_channel_open{channel_type = "forwarded-tcpip",
             {ok, {ConnectToHost,ConnectToPort}} ->
                 case gen_tcp:connect(ConnectToHost, ConnectToPort, [{active,false}, binary]) of
                     {ok,Sock} ->
-                        {ok,Pid} = ssh_subsystem_sup:start_channel(client, SubSysSup, self(),
+                        {ok,Pid} = ssh_connection_sup:start_channel(client, ConnectionSup, self(),
                                                                    ssh_tcpip_forward_client, ChId,
                                                                    [Sock], undefined, Options),
                         ssh_client_channel:cache_update(Cache,
@@ -942,12 +972,22 @@ handle_msg(#ssh_msg_channel_open{channel_type = "direct-tcpip",
                        suggest_window_size = WinSz,
                        suggest_packet_size = PktSz,
                        options = Options,
-                       sub_system_supervisor = SubSysSup
+                       connection_supervisor = ConnectionSup
                       } = C,
 	   server, _SSH) ->
+    Allowed = case ?GET_OPT(tcpip_tunnel_in, Options) of
+                  T when is_boolean(T) -> T;
+                  AllowedFun when is_function(AllowedFun, 2) ->
+                      AllowedFun(binary_to_list(HostToConnect), PortToConnect)
+              end,
     {ReplyMsg, NextChId} =
-        case ?GET_OPT(tcpip_tunnel_in, Options) of
-            %% May add more to the option, like allowed ip/port pairs to connect to
+        case Allowed of
+            denied ->
+                {channel_open_failure_msg(RemoteId,
+                                          ?SSH_OPEN_ADMINISTRATIVELY_PROHIBITED,
+                                          "Not allowed", "en"),
+                 ChId};
+
             false ->
                 {channel_open_failure_msg(RemoteId, 
                                           ?SSH_OPEN_CONNECT_FAILED,
@@ -958,7 +998,7 @@ handle_msg(#ssh_msg_channel_open{channel_type = "direct-tcpip",
                 case gen_tcp:connect(binary_to_list(HostToConnect), PortToConnect,
                                      [{active,false}, binary]) of
                     {ok,Sock} ->
-                        {ok,Pid} = ssh_subsystem_sup:start_channel(server, SubSysSup, self(),
+                        {ok,Pid} = ssh_connection_sup:start_channel(server, ConnectionSup, self(),
                                                                    ssh_tcpip_forward_srv, ChId,
                                                                    [Sock], undefined, Options),
                         ssh_client_channel:cache_update(Cache,
@@ -1017,21 +1057,25 @@ handle_msg(#ssh_msg_channel_request{recipient_channel = ChannelId,
 handle_msg(#ssh_msg_channel_request{recipient_channel = ChannelId,
 				    request_type = "exit-signal",
 				    want_reply = false,
-				    data = Data},  
+				    data = Data},
            #connection{channel_cache = Cache} = Connection0, _, _SSH) ->
     <<?DEC_BIN(SigName, _SigLen),
-      ?BOOLEAN(_Core), 
+      ?BOOLEAN(_Core),
       ?DEC_BIN(Err, _ErrLen),
       ?DEC_BIN(Lang, _LangLen)>> = Data,
-    Channel = ssh_client_channel:cache_lookup(Cache, ChannelId),
-    RemoteId =  Channel#channel.remote_id,
-    {Reply, Connection} =  reply_msg(Channel, Connection0, 
-				     {exit_signal, ChannelId,
-				      binary_to_list(SigName),
-				      binary_to_list(Err),
-				      binary_to_list(Lang)}),
-    CloseMsg = channel_close_msg(RemoteId),
-    {[{connection_reply, CloseMsg}|Reply], Connection};
+    case ssh_client_channel:cache_lookup(Cache, ChannelId) of
+        #channel{remote_id = RemoteId} = Channel ->
+            {Reply, Connection} =  reply_msg(Channel, Connection0,
+                                             {exit_signal, ChannelId,
+                                              binary_to_list(SigName),
+                                              binary_to_list(Err),
+                                              binary_to_list(Lang)}),
+            ChannelCloseMsg = channel_close_msg(RemoteId),
+            {[{connection_reply, ChannelCloseMsg}|Reply], Connection};
+        _ ->
+            %% Channel already closed by peer
+            {[], Connection0}
+    end;
 
 handle_msg(#ssh_msg_channel_request{recipient_channel = ChannelId,
 				    request_type = "xon-xoff",
@@ -1190,8 +1234,8 @@ handle_msg(#ssh_msg_global_request{name = <<"tcpip-forward">>,
             {[{connection_reply, request_failure_msg()}], Connection};
 
         true ->
-            SubSysSup = ?GET_INTERNAL_OPT(subsystem_sup, Opts),
-            FwdSup = ssh_subsystem_sup:tcpip_fwd_supervisor(SubSysSup),
+            ConnectionSup = ?GET_INTERNAL_OPT(connection_sup, Opts),
+            FwdSup = ssh_connection_sup:tcpip_fwd_supervisor(ConnectionSup),
             ConnPid = self(),
             case ssh_tcpip_forward_acceptor:supervised_start(FwdSup,
                                                              {ListenAddrStr, ListenPort},
@@ -1244,12 +1288,7 @@ handle_msg(#ssh_msg_request_success{data = Data},
 	   #connection{requests = [{_, From, Fun} | Rest]} = Connection0, _, _SSH) ->
     Connection = Fun({success,Data}, Connection0),
     {[{channel_request_reply, From, {success, Data}}],
-     Connection#connection{requests = Rest}};
-
-handle_msg(#ssh_msg_disconnect{code = Code,
-			       description = Description},
-	   Connection, _, _SSH) ->
-    {disconnect, {Code, Description}, handle_stop(Connection)}.
+     Connection#connection{requests = Rest}}.
 
 
 %%%----------------------------------------------------------------
@@ -1421,22 +1460,22 @@ setup_session(#connection{channel_cache = Cache,
 start_cli(#connection{options = Options, 
 		      cli_spec = CliSpec,
 		      exec = Exec,
-		      sub_system_supervisor = SubSysSup}, ChannelId) ->
+		      connection_supervisor = ConnectionSup}, ChannelId) ->
     case CliSpec of
         no_cli ->
             {error, cli_disabled};
         {CbModule, Args} ->
-            ssh_subsystem_sup:start_channel(server, SubSysSup, self(), CbModule, ChannelId, Args, Exec, Options)
+            ssh_connection_sup:start_channel(server, ConnectionSup, self(), CbModule, ChannelId, Args, Exec, Options)
     end.
 
 
 start_subsystem(BinName, #connection{options = Options,
-                                     sub_system_supervisor = SubSysSup},
+                                     connection_supervisor = ConnectionSup},
 	       #channel{local_id = ChannelId}, _ReplyMsg) ->
     Name = binary_to_list(BinName),
     case check_subsystem(Name, Options) of
 	{Callback, Opts} when is_atom(Callback), Callback =/= none ->
-            ssh_subsystem_sup:start_channel(server, SubSysSup, self(), Callback, ChannelId, Opts, undefined, Options);
+            ssh_connection_sup:start_channel(server, ConnectionSup, self(), Callback, ChannelId, Opts, undefined, Options);
         {none, _} ->
             {error, bad_subsystem};
 	{_, _} ->
@@ -1549,9 +1588,9 @@ flow_control(_,_,_) ->
 %%% Pseudo terminal stuff
 %%% 
 
-pty_req(ConnectionHandler, Channel, Term, Width, Height,
+pty_req(ConnectionRef, Channel, Term, Width, Height,
 	 PixWidth, PixHeight, PtyOpts, TimeOut) ->
-    ssh_connection_handler:request(ConnectionHandler,
+    ssh_connection_handler:request(ConnectionRef,
 				   Channel, "pty-req", true,
 				   [?string(Term),
 				    ?uint32(Width), ?uint32(Height),
@@ -1819,7 +1858,6 @@ channel_data_reply_msg(ChannelId, Connection, DataType, Data) ->
 	    WantedSize = Size - byte_size(Data),
 	    ssh_client_channel:cache_update(Connection#connection.channel_cache, 
                                      Channel#channel{recv_window_size = WantedSize}),
-            adjust_window(self(), ChannelId, byte_size(Data)),
             reply_msg(Channel, Connection, {data, ChannelId, DataType, Data});
 	undefined ->
 	    {[], Connection}
@@ -1861,14 +1899,14 @@ request_reply_or_data(#channel{local_id = ChannelId, user = ChannelPid},
 
 %%%----------------------------------------------------------------
 -doc false.
-send_environment_vars(ConnectionHandler, Channel, VarNames) ->
+send_environment_vars(ConnectionRef, Channel, VarNames) ->
     lists:foldl(
       fun(Var, success) ->
               case os:getenv(Var) of
                   false ->
                       success;
                   Value ->
-                      setenv(ConnectionHandler, Channel, false,
+                      setenv(ConnectionRef, Channel, false,
                              Var, Value, infinity)
               end
       end, success, VarNames).
