@@ -1,18 +1,21 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2006-2015. All Rights Reserved.
+%% SPDX-License-Identifier: Apache-2.0
 %%
-%% The contents of this file are subject to the Erlang Public License,
-%% Version 1.1, (the "License"); you may not use this file except in
-%% compliance with the License. You should have received a copy of the
-%% Erlang Public License along with this software. If not, it can be
-%% retrieved online at http://www.erlang.org/.
+%% Copyright Ericsson AB 2006-2025. All Rights Reserved.
 %%
-%% Software distributed under the License is distributed on an "AS IS"
-%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
-%% the License for the specific language governing rights and limitations
-%% under the License.
+%% Licensed under the Apache License, Version 2.0 (the "License");
+%% you may not use this file except in compliance with the License.
+%% You may obtain a copy of the License at
+%%
+%%     http://www.apache.org/licenses/LICENSE-2.0
+%%
+%% Unless required by applicable law or agreed to in writing, software
+%% distributed under the License is distributed on an "AS IS" BASIS,
+%% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+%% See the License for the specific language governing permissions and
+%% limitations under the License.
 %%
 %% %CopyrightEnd%
 %%
@@ -20,19 +23,54 @@
 %%
 -module(ssh_sftpd_SUITE).
 
-%% Note: This directive should only be used in test suites.
--compile(export_all).
+-export([
+         suite/0,
+         all/0,
+         groups/0,
+         init_per_suite/1,
+         end_per_suite/1,
+         init_per_group/2,
+         end_per_group/2,
+         init_per_testcase/2,
+         end_per_testcase/2
+        ]).
+
+-export([
+         access_outside_root/1,
+         links/1,
+         mk_rm_dir/1,
+         open_close_dir/1,
+         open_close_file/1,
+         open_file_dir_v5/1,
+         open_file_dir_v6/1,
+         read_dir/1,
+         read_file/1,
+         real_path/1,
+         relative_path/1,
+         relpath/1,
+         remove_file/1,
+         rename_file/1,
+         retrieve_attributes/1,
+         root_with_cwd/1,
+         set_attributes/1,
+         sshd_read_file/1,
+         ver3_open_flags/1,
+         ver3_rename/1,
+         ver6_basic/1,
+         write_file/1
+        ]).
 
 -include_lib("common_test/include/ct.hrl").
 -include_lib("kernel/include/file.hrl").
 -include("ssh_xfer.hrl").
 -include("ssh.hrl").
+-include("ssh_test_lib.hrl").
 
 -define(USER, "Alladin").
 -define(PASSWD, "Sesame").
--define(XFER_PACKET_SIZE, 32768).
--define(XFER_WINDOW_SIZE, 4*?XFER_PACKET_SIZE).
--define(TIMEOUT, 10000).
+%% -define(XFER_PACKET_SIZE, 32768).
+%% -define(XFER_WINDOW_SIZE, 4*?XFER_PACKET_SIZE).
+-define(SSH_TIMEOUT, 5000).
 -define(REG_ATTERS, <<0,0,0,0,1>>).
 -define(UNIX_EPOCH,  62167219200).
 
@@ -43,6 +81,9 @@
 %% Common Test interface functions -----------------------------------
 %%--------------------------------------------------------------------
 
+suite() ->
+    [{timetrap,{seconds,20}}].
+
 all() -> 
     [open_close_file, 
      open_close_dir, 
@@ -60,7 +101,12 @@ all() ->
      ver3_open_flags,
      relpath, 
      sshd_read_file,
-     ver6_basic].
+     ver6_basic,
+     access_outside_root,
+     root_with_cwd,
+     relative_path,
+     open_file_dir_v5,
+     open_file_dir_v6].
 
 groups() -> 
     [].
@@ -68,27 +114,22 @@ groups() ->
 %%--------------------------------------------------------------------
 
 init_per_suite(Config) ->
-    case (catch crypto:start()) of
-	ok ->
-	    DataDir = ?config(data_dir, Config),	    
-	    PrivDir = ?config(priv_dir, Config),
-	    ssh_test_lib:setup_dsa(DataDir, PrivDir),
-	    %% to make sure we don't use public-key-auth
-	    %% this should be tested by other test suites
-	    UserDir = filename:join(?config(priv_dir, Config), nopubkey), 
-	    file:make_dir(UserDir),  
-	    Config;
-	_ ->
-	    {skip,"Could not start crypto!"}
-    end.
+    ?CHECK_CRYPTO(
+       begin
+           ssh:start(),
+           ct:log("Pub keys setup for: ~p",
+                  [ssh_test_lib:setup_all_user_host_keys(Config)]),
+	   %% to make sure we don't use public-key-auth
+	   %% this should be tested by other test suites
+	   UserDir = filename:join(proplists:get_value(priv_dir, Config), nopubkey), 
+	   file:make_dir(UserDir),  
+	   Config
+       end).
 
 end_per_suite(Config) ->
-    SysDir = ?config(priv_dir, Config),
-    ssh_test_lib:clean_dsa(SysDir),
-    UserDir = filename:join(?config(priv_dir, Config), nopubkey),
+    UserDir = filename:join(proplists:get_value(priv_dir, Config), nopubkey),
     file:del_dir(UserDir),
-    ssh:stop(),
-    crypto:stop().
+    ssh:stop().
 
 %%--------------------------------------------------------------------
 
@@ -103,11 +144,10 @@ end_per_group(_GroupName, Config) ->
 init_per_testcase(TestCase, Config) ->
     ssh:start(),
     prep(Config),
-    PrivDir = ?config(priv_dir, Config),
+    PrivDir = proplists:get_value(priv_dir, Config),
     ClientUserDir = filename:join(PrivDir, nopubkey),
-    SystemDir = filename:join(?config(priv_dir, Config), system),
+    SystemDir = filename:join(proplists:get_value(priv_dir, Config), system),
 
-    Port = ssh_test_lib:inet_port(node()),
     Options = [{system_dir, SystemDir},
 	       {user_dir, PrivDir},
 	       {user_passwords,[{?USER, ?PASSWD}]},
@@ -115,23 +155,49 @@ init_per_testcase(TestCase, Config) ->
     {ok, Sftpd} = case TestCase of
 		      ver6_basic ->
 			  SubSystems = [ssh_sftpd:subsystem_spec([{sftpd_vsn, 6}])],
-			  ssh:daemon(Port, [{subsystems, SubSystems}|Options]);
+			  ssh:daemon(0, [{subsystems, SubSystems}|Options]);
+                      access_outside_root ->
+                          %% Build RootDir/access_outside_root/a/b and set Root and CWD
+                          BaseDir = filename:join(PrivDir, access_outside_root),
+                          RootDir = filename:join(BaseDir, a),
+                          CWD     = filename:join(RootDir, b),
+                          %% Make the directory chain:
+                          ok = filelib:ensure_dir(filename:join(CWD, tmp)),
+                          SubSystems = [ssh_sftpd:subsystem_spec([{root, RootDir},
+                                                                  {cwd, CWD}])],
+                          ssh:daemon(0, [{subsystems, SubSystems}|Options]);
+		      root_with_cwd ->
+			  RootDir = filename:join(PrivDir, root_with_cwd),
+			  CWD     = filename:join(RootDir, home),
+			  SubSystems = [ssh_sftpd:subsystem_spec([{root, RootDir}, {cwd, CWD}])],
+			  ssh:daemon(0, [{subsystems, SubSystems}|Options]);
+		      relative_path ->
+			  SubSystems = [ssh_sftpd:subsystem_spec([{cwd, PrivDir}])],
+			  ssh:daemon(0, [{subsystems, SubSystems}|Options]);
+		      open_file_dir_v5 ->
+			  SubSystems = [ssh_sftpd:subsystem_spec([{cwd, PrivDir}])],
+			  ssh:daemon(0, [{subsystems, SubSystems}|Options]);
+		      open_file_dir_v6 ->
+			  SubSystems = [ssh_sftpd:subsystem_spec([{cwd, PrivDir},
+								  {sftpd_vsn, 6}])],
+			  ssh:daemon(0, [{subsystems, SubSystems}|Options]);
 		      _ ->
 			  SubSystems = [ssh_sftpd:subsystem_spec([])],
-			  ssh:daemon(Port, [{subsystems, SubSystems}|Options])
+			  ssh:daemon(0, [{subsystems, SubSystems}|Options])
 		  end,
+
+    Port = ssh_test_lib:daemon_port(Sftpd),
     
     Cm = ssh_test_lib:connect(Port,
 			      [{user_dir, ClientUserDir},
 			       {user, ?USER}, {password, ?PASSWD},
 			       {user_interaction, false},
-			       {silently_accept_hosts, true},
-			       {pwdfun, fun(_,_) -> true end}]),
+			       {silently_accept_hosts, true}]),
     {ok, Channel} =
 	ssh_connection:session_channel(Cm, ?XFER_WINDOW_SIZE,
-				       ?XFER_PACKET_SIZE, ?TIMEOUT),
+				       ?XFER_PACKET_SIZE, ?SSH_TIMEOUT),
     
-    success = ssh_connection:subsystem(Cm, Channel, "sftp", ?TIMEOUT),
+    success = ssh_connection:subsystem(Cm, Channel, "sftp", ?SSH_TIMEOUT),
 
     ProtocolVer = case atom_to_list(TestCase) of
 		      "ver3_" ++ _ ->
@@ -150,13 +216,13 @@ init_per_testcase(TestCase, Config) ->
     {ok, <<?SSH_FXP_VERSION, ?UINT32(Version), _Ext/binary>>, _}
 	= reply(Cm, Channel),
 
-    ct:pal("Client: ~p Server ~p~n", [ProtocolVer, Version]),
+    ct:log("Client: ~p Server ~p~n", [ProtocolVer, Version]),
 
     [{sftp, {Cm, Channel}}, {sftpd, Sftpd }| Config].
 
 end_per_testcase(_TestCase, Config) ->
-    ssh_sftpd:stop(?config(sftpd, Config)),
-    {Cm, Channel} = ?config(sftp, Config),
+    catch ssh:stop_daemon(proplists:get_value(sftpd, Config)),
+    {Cm, Channel} = proplists:get_value(sftp, Config),
     ssh_connection:close(Cm, Channel),
     ssh:close(Cm),
     ssh:stop().
@@ -164,12 +230,10 @@ end_per_testcase(_TestCase, Config) ->
 %%--------------------------------------------------------------------
 %% Test Cases --------------------------------------------------------
 %%--------------------------------------------------------------------
-open_close_file() ->
-    [{doc, "Test SSH_FXP_OPEN and SSH_FXP_CLOSE commands"}].
 open_close_file(Config) when is_list(Config) ->
-    PrivDir =  ?config(priv_dir, Config),
+    PrivDir =  proplists:get_value(priv_dir, Config),
     FileName = filename:join(PrivDir, "test.txt"),
-    {Cm, Channel} = ?config(sftp, Config),
+    {Cm, Channel} = proplists:get_value(sftp, Config),
     ReqId = 0,
 
     {ok, <<?SSH_FXP_HANDLE, ?UINT32(ReqId), Handle/binary>>, _} =
@@ -194,12 +258,10 @@ open_close_file(Config) when is_list(Config) ->
 		  ?ACE4_READ_DATA  bor ?ACE4_READ_ATTRIBUTES,
 		  ?SSH_FXF_OPEN_EXISTING).
 
-ver3_open_flags() ->
-    [{doc, "Test open flags"}].
 ver3_open_flags(Config) when is_list(Config) ->
-    PrivDir =  ?config(priv_dir, Config),
+    PrivDir =  proplists:get_value(priv_dir, Config),
     FileName = filename:join(PrivDir, "not_exist.txt"),
-    {Cm, Channel} = ?config(sftp, Config),
+    {Cm, Channel} = proplists:get_value(sftp, Config),
     ReqId = 0,
     
     {ok, <<?SSH_FXP_HANDLE, ?UINT32(ReqId), Handle/binary>>, _} =
@@ -228,11 +290,9 @@ ver3_open_flags(Config) when is_list(Config) ->
 						       Cm, Channel).
     
 %%--------------------------------------------------------------------
-open_close_dir() ->
-    [{doc,"Test SSH_FXP_OPENDIR and SSH_FXP_CLOSE commands"}].
 open_close_dir(Config) when is_list(Config) ->
-    PrivDir = ?config(priv_dir, Config),
-    {Cm, Channel} = ?config(sftp, Config),
+    PrivDir = proplists:get_value(priv_dir, Config),
+    {Cm, Channel} = proplists:get_value(sftp, Config),
     FileName = filename:join(PrivDir, "test.txt"),
     ReqId = 0,
 
@@ -255,14 +315,12 @@ open_close_dir(Config) when is_list(Config) ->
     end.
 
 %%--------------------------------------------------------------------
-read_file() ->
-    [{doc, "Test SSH_FXP_READ command"}].
 read_file(Config) when is_list(Config) ->
-    PrivDir =  ?config(priv_dir, Config),
+    PrivDir =  proplists:get_value(priv_dir, Config),
     FileName = filename:join(PrivDir, "test.txt"),
 
     ReqId = 0,
-    {Cm, Channel} = ?config(sftp, Config),
+    {Cm, Channel} = proplists:get_value(sftp, Config),
 
     {ok, <<?SSH_FXP_HANDLE, ?UINT32(ReqId), Handle/binary>>, _} =
 	open_file(FileName, Cm, Channel, ReqId,
@@ -278,25 +336,21 @@ read_file(Config) when is_list(Config) ->
     {ok, Data} = file:read_file(FileName).
 
 %%--------------------------------------------------------------------
-read_dir() ->
-    [{doc,"Test SSH_FXP_READDIR command"}].
 read_dir(Config) when is_list(Config) ->
-    PrivDir = ?config(priv_dir, Config),
-    {Cm, Channel} = ?config(sftp, Config),
+    PrivDir = proplists:get_value(priv_dir, Config),
+    {Cm, Channel} = proplists:get_value(sftp, Config),
     ReqId = 0,
     {ok, <<?SSH_FXP_HANDLE, ?UINT32(ReqId), Handle/binary>>, _} =
 	open_dir(PrivDir, Cm, Channel, ReqId),
     ok = read_dir(Handle, Cm, Channel, ReqId).
 
 %%--------------------------------------------------------------------
-write_file() ->
-    [{doc, "Test SSH_FXP_WRITE command"}].
 write_file(Config) when is_list(Config) ->
-    PrivDir =  ?config(priv_dir, Config),
+    PrivDir =  proplists:get_value(priv_dir, Config),
     FileName = filename:join(PrivDir, "test.txt"),
 
     ReqId = 0,
-    {Cm, Channel} = ?config(sftp, Config),
+    {Cm, Channel} = proplists:get_value(sftp, Config),
 
     {ok, <<?SSH_FXP_HANDLE, ?UINT32(ReqId), Handle/binary>>, _} =
 	open_file(FileName, Cm, Channel, ReqId,
@@ -313,13 +367,11 @@ write_file(Config) when is_list(Config) ->
     {ok, Data} = file:read_file(FileName).
 
 %%--------------------------------------------------------------------
-remove_file() ->
-    [{doc, "Test SSH_FXP_REMOVE command"}].
 remove_file(Config) when is_list(Config) ->
-    PrivDir =  ?config(priv_dir, Config),
+    PrivDir =  proplists:get_value(priv_dir, Config),
     FileName = filename:join(PrivDir, "test.txt"),
     ReqId = 0,
-    {Cm, Channel} = ?config(sftp, Config),
+    {Cm, Channel} = proplists:get_value(sftp, Config),
 
     {ok, <<?SSH_FXP_STATUS, ?UINT32(ReqId),
 	   ?UINT32(?SSH_FX_OK), _/binary>>, _} =
@@ -334,14 +386,12 @@ remove_file(Config) when is_list(Config) ->
 	remove(PrivDir, Cm, Channel, NewReqId).
 
 %%--------------------------------------------------------------------
-rename_file() ->
-    [{doc, "Test SSH_FXP_RENAME command"}].
 rename_file(Config) when is_list(Config) ->
-    PrivDir =  ?config(priv_dir, Config),
+    PrivDir =  proplists:get_value(priv_dir, Config),
     FileName = filename:join(PrivDir, "test.txt"),
     NewFileName = filename:join(PrivDir, "test1.txt"),
     ReqId = 0,
-    {Cm, Channel} = ?config(sftp, Config),
+    {Cm, Channel} = proplists:get_value(sftp, Config),
 
     {ok, <<?SSH_FXP_STATUS, ?UINT32(ReqId),
 	  ?UINT32(?SSH_FX_OK), _/binary>>, _} =
@@ -357,7 +407,7 @@ rename_file(Config) when is_list(Config) ->
     NewReqId1 = NewReqId + 1,
     file:copy(FileName, NewFileName),
 
-    %% No owerwrite
+    %% No overwrite
     {ok, <<?SSH_FXP_STATUS, ?UINT32(NewReqId1),
 	  ?UINT32(?SSH_FX_FILE_ALREADY_EXISTS), _/binary>>, _} =
 	rename(FileName, NewFileName, Cm, Channel, NewReqId1, 6,
@@ -371,11 +421,9 @@ rename_file(Config) when is_list(Config) ->
 	       ?SSH_FXP_RENAME_ATOMIC).
 
 %%--------------------------------------------------------------------
-mk_rm_dir() ->
-    [{doc, "Test SSH_FXP_MKDIR and SSH_FXP_RMDIR command"}].
 mk_rm_dir(Config) when is_list(Config) ->
-    PrivDir = ?config(priv_dir, Config),
-    {Cm, Channel} = ?config(sftp, Config),
+    PrivDir = proplists:get_value(priv_dir, Config),
+    {Cm, Channel} = proplists:get_value(sftp, Config),
     DirName = filename:join(PrivDir, "test"),
     ReqId = 0,
     {ok, <<?SSH_FXP_STATUS, ?UINT32(ReqId), ?UINT32(?SSH_FX_OK),
@@ -394,16 +442,14 @@ mk_rm_dir(Config) when is_list(Config) ->
 	    _/binary>>, _} = rmdir(DirName, Cm, Channel, NewReqId2).
 
 %%--------------------------------------------------------------------
-real_path() ->
-    [{doc, "Test SSH_FXP_REALPATH command"}].
 real_path(Config) when is_list(Config) ->
     case os:type() of
 	{win32, _} ->
 	    {skip,  "Not a relevant test on windows"};
 	_ ->
 	    ReqId = 0,
-	    {Cm, Channel} = ?config(sftp, Config),
-	    PrivDir = ?config(priv_dir, Config),
+	    {Cm, Channel} = proplists:get_value(sftp, Config),
+	    PrivDir = proplists:get_value(priv_dir, Config),
 	    TestDir = filename:join(PrivDir, "ssh_test"),
 	    ok = file:make_dir(TestDir),
 
@@ -416,7 +462,7 @@ real_path(Config) when is_list(Config) ->
 	    RealPath = filename:absname(binary_to_list(Path)),
 	    AbsPrivDir = filename:absname(PrivDir),
 
-	    ct:pal("Path: ~p PrivDir: ~p~n", [RealPath, AbsPrivDir]),
+	    ct:log("Path: ~p PrivDir: ~p~n", [RealPath, AbsPrivDir]),
 
 	    true = RealPath == AbsPrivDir
     end.
@@ -428,8 +474,8 @@ links(Config) when is_list(Config) ->
 	    {skip, "Links are not fully supported by windows"};
 	_ ->
 	    ReqId = 0,
-	    {Cm, Channel} = ?config(sftp, Config),
-	    PrivDir =  ?config(priv_dir, Config),
+	    {Cm, Channel} = proplists:get_value(sftp, Config),
+	    PrivDir =  proplists:get_value(priv_dir, Config),
 	    FileName = filename:join(PrivDir, "test.txt"),
 	    LinkFileName = filename:join(PrivDir, "link_test.txt"),
 
@@ -445,17 +491,15 @@ links(Config) when is_list(Config) ->
 
 	    true = binary_to_list(Path) == FileName,
 
-	    ct:pal("Path: ~p~n", [binary_to_list(Path)])
+	    ct:log("Path: ~p~n", [binary_to_list(Path)])
     end.
 
 %%--------------------------------------------------------------------
-retrieve_attributes() ->
-    [{"Test SSH_FXP_STAT, SSH_FXP_LSTAT AND SSH_FXP_FSTAT commands"}].
 retrieve_attributes(Config) when is_list(Config) ->
-    PrivDir =  ?config(priv_dir, Config),
+    PrivDir =  proplists:get_value(priv_dir, Config),
     FileName = filename:join(PrivDir, "test.txt"),
     ReqId = 0,
-    {Cm, Channel} = ?config(sftp, Config),
+    {Cm, Channel} = proplists:get_value(sftp, Config),
 
     {ok, FileInfo} = file:read_file_info(FileName),
 
@@ -514,17 +558,15 @@ retrieve_attributes(Config) when is_list(Config) ->
 		  end, AttrValues).
 
 %%--------------------------------------------------------------------
-set_attributes() ->
-    [{doc, "Test SSH_FXP_SETSTAT AND SSH_FXP_FSETSTAT commands"}].
 set_attributes(Config) when is_list(Config) ->
     case os:type() of
 	{win32, _} ->
 	    {skip,  "Known error bug in erts file:read_file_info"};
 	_ ->
-	    PrivDir =  ?config(priv_dir, Config),
+	    PrivDir =  proplists:get_value(priv_dir, Config),
 	    FileName = filename:join(PrivDir, "test.txt"),
 	    ReqId = 0,
-	    {Cm, Channel} = ?config(sftp, Config),
+	    {Cm, Channel} = proplists:get_value(sftp, Config),
 
 	    {ok, FileInfo} = file:read_file_info(FileName),
 
@@ -546,10 +588,10 @@ set_attributes(Config) when is_list(Config) ->
 	    %% Can not test that NewPermissions = Permissions as
 	    %% on Unix platforms, other bits than those listed in the
 	    %% API may be set.
-	    ct:pal("Org: ~p New: ~p~n", [OrigPermissions, NewPermissions]),
+	    ct:log("Org: ~p New: ~p~n", [OrigPermissions, NewPermissions]),
 	    true = OrigPermissions =/= NewPermissions,
 
-	    ct:pal("Try to open the file"),
+	    ct:log("Try to open the file"),
 	    NewReqId = 2,
 	    {ok, <<?SSH_FXP_HANDLE, ?UINT32(NewReqId), Handle/binary>>, _} =
 		open_file(FileName, Cm, Channel, NewReqId,
@@ -561,7 +603,7 @@ set_attributes(Config) when is_list(Config) ->
 
 	    NewReqId1 = 3,
 
-	    ct:pal("Set original permissions on the now open file"),
+	    ct:log("Set original permissions on the now open file"),
 
 	    {ok, <<?SSH_FXP_STATUS, ?UINT32(NewReqId1),
 		   ?UINT32(?SSH_FX_OK), _/binary>>, _} =
@@ -572,25 +614,21 @@ set_attributes(Config) when is_list(Config) ->
     end.
 
 %%--------------------------------------------------------------------
-ver3_rename() ->
-    [{doc, "Test that ver3 rename message is handled OTP 6352"}].
 ver3_rename(Config) when is_list(Config) ->
-    PrivDir =  ?config(priv_dir, Config),
+    PrivDir =  proplists:get_value(priv_dir, Config),
     FileName = filename:join(PrivDir, "test.txt"),
     NewFileName = filename:join(PrivDir, "test1.txt"),
     ReqId = 0,
-    {Cm, Channel} = ?config(sftp, Config),
+    {Cm, Channel} = proplists:get_value(sftp, Config),
 
     {ok, <<?SSH_FXP_STATUS, ?UINT32(ReqId),
 	  ?UINT32(?SSH_FX_OK), _/binary>>, _} =
 	rename(FileName, NewFileName, Cm, Channel, ReqId, 3, 0).
 
 %%--------------------------------------------------------------------
-relpath() ->
-    [{doc, "Check that realpath works ok seq10670"}].
 relpath(Config) when is_list(Config) ->
     ReqId = 0,
-    {Cm, Channel} = ?config(sftp, Config),
+    {Cm, Channel} = proplists:get_value(sftp, Config),
 
     case os:type() of
 	{win32, _} ->
@@ -609,14 +647,12 @@ relpath(Config) when is_list(Config) ->
     end.
 
 %%--------------------------------------------------------------------
-sshd_read_file() ->
-    [{doc,"Test SSH_FXP_READ command, using sshd-server"}].
 sshd_read_file(Config) when is_list(Config) ->
-    PrivDir =  ?config(priv_dir, Config),
+    PrivDir =  proplists:get_value(priv_dir, Config),
     FileName = filename:join(PrivDir, "test.txt"),
 
     ReqId = 0,
-    {Cm, Channel} = ?config(sftp, Config),
+    {Cm, Channel} = proplists:get_value(sftp, Config),
 
     {ok, <<?SSH_FXP_HANDLE, ?UINT32(ReqId), Handle/binary>>, _} =
 	open_file(FileName, Cm, Channel, ReqId,
@@ -631,23 +667,138 @@ sshd_read_file(Config) when is_list(Config) ->
 
     {ok, Data} = file:read_file(FileName).
 %%--------------------------------------------------------------------
-ver6_basic() ->
-    [{doc, "Test SFTP Version 6"}].
 ver6_basic(Config) when is_list(Config) ->
-    PrivDir =  ?config(priv_dir, Config),
+    PrivDir =  proplists:get_value(priv_dir, Config),
     %FileName = filename:join(PrivDir, "test.txt"),
-    {Cm, Channel} = ?config(sftp, Config),
+    {Cm, Channel} = proplists:get_value(sftp, Config),
     ReqId = 0,
     {ok, <<?SSH_FXP_STATUS, ?UINT32(ReqId),  % Ver 6 we have 5
 	   ?UINT32(?SSH_FX_FILE_IS_A_DIRECTORY), _/binary>>, _} =
 	open_file(PrivDir, Cm, Channel, ReqId,
 		  ?ACE4_READ_DATA  bor ?ACE4_READ_ATTRIBUTES,
 		  ?SSH_FXF_OPEN_EXISTING).
+
+%%--------------------------------------------------------------------
+access_outside_root(Config) when is_list(Config) ->
+    PrivDir  =  proplists:get_value(priv_dir, Config),
+    BaseDir  = filename:join(PrivDir, access_outside_root),
+    %% A file outside the tree below RootDir which is BaseDir/a
+    %% Make the file  BaseDir/bad :
+    BadFilePath = filename:join([BaseDir, bad]),
+    ok = file:write_file(BadFilePath, <<>>),
+    {Cm, Channel} = proplists:get_value(sftp, Config),
+    %% Try to access a file parallel to the RootDir:
+    try_access("/../bad",   Cm, Channel, 0),
+    %% Try to access the same file via the CWD which is /b relative to the RootDir:
+    try_access("../../bad", Cm, Channel, 1).
+
+
+try_access(Path, Cm, Channel, ReqId) ->
+    Return = 
+        open_file(Path, Cm, Channel, ReqId, 
+                  ?ACE4_READ_DATA bor ?ACE4_READ_ATTRIBUTES,
+                  ?SSH_FXF_OPEN_EXISTING),
+    ct:log("Try open ~p -> ~p",[Path,Return]),
+    case Return of
+        {ok, <<?SSH_FXP_HANDLE, ?UINT32(ReqId), _Handle0/binary>>, _} ->
+            ct:fail("Could open a file outside the root tree!");
+        {ok, <<?SSH_FXP_STATUS, ?UINT32(ReqId), ?UINT32(Code), Rest/binary>>, <<>>} ->
+            case Code of
+                ?SSH_FX_FILE_IS_A_DIRECTORY ->
+                    ct:log("Got the expected SSH_FX_FILE_IS_A_DIRECTORY status",[]),
+                    ok;
+                ?SSH_FX_FAILURE ->
+                    ct:log("Got the expected SSH_FX_FAILURE status",[]),
+                    ok;
+                _ ->
+                    case Rest of
+                        <<?UINT32(Len), Txt:Len/binary, _/binary>> ->
+                            ct:fail("Got unexpected SSH_FX_code: ~p (~p)",[Code,Txt]);
+                        _ ->
+                            ct:fail("Got unexpected SSH_FX_code: ~p",[Code])
+                    end
+            end;
+        _ ->
+            ct:fail("Completely unexpected return: ~p", [Return])
+    end.
+
+%%--------------------------------------------------------------------
+root_with_cwd(Config) when is_list(Config) ->
+    PrivDir =  proplists:get_value(priv_dir, Config),
+    RootDir = filename:join(PrivDir, root_with_cwd),
+    CWD     = filename:join(RootDir, home),
+    FileName = "root_with_cwd.txt",
+    FilePath = filename:join(CWD, FileName),
+    ok = filelib:ensure_dir(FilePath),
+    ok = file:write_file(FilePath ++ "0", <<>>),
+    ok = file:write_file(FilePath ++ "1", <<>>),
+    ok = file:write_file(FilePath ++ "2", <<>>),
+    {Cm, Channel} = proplists:get_value(sftp, Config),
+    ReqId0 = 0,
+    {ok, <<?SSH_FXP_HANDLE, ?UINT32(ReqId0), _Handle0/binary>>, _} =
+	open_file(FileName ++ "0", Cm, Channel, ReqId0,
+		  ?ACE4_READ_DATA  bor ?ACE4_READ_ATTRIBUTES,
+		  ?SSH_FXF_OPEN_EXISTING),
+    ReqId1 = 1,
+    {ok, <<?SSH_FXP_HANDLE, ?UINT32(ReqId1), _Handle1/binary>>, _} =
+	open_file("./" ++ FileName ++ "1", Cm, Channel, ReqId1,
+		  ?ACE4_READ_DATA  bor ?ACE4_READ_ATTRIBUTES,
+		  ?SSH_FXF_OPEN_EXISTING),
+    ReqId2 = 2,
+    {ok, <<?SSH_FXP_HANDLE, ?UINT32(ReqId2), _Handle2/binary>>, _} =
+	open_file("/home/" ++ FileName ++ "2", Cm, Channel, ReqId2,
+		  ?ACE4_READ_DATA  bor ?ACE4_READ_ATTRIBUTES,
+		  ?SSH_FXF_OPEN_EXISTING).
+
+%%--------------------------------------------------------------------
+relative_path(Config) when is_list(Config) ->
+    PrivDir =  proplists:get_value(priv_dir, Config),
+    FileName = "test_relative_path.txt",
+    FilePath = filename:join(PrivDir, FileName),
+    ok = filelib:ensure_dir(FilePath),
+    ok = file:write_file(FilePath, <<>>),
+    {Cm, Channel} = proplists:get_value(sftp, Config),
+    ReqId = 0,
+    {ok, <<?SSH_FXP_HANDLE, ?UINT32(ReqId), _Handle/binary>>, _} =
+        open_file(FileName, Cm, Channel, ReqId,
+                  ?ACE4_READ_DATA  bor ?ACE4_READ_ATTRIBUTES,
+                  ?SSH_FXF_OPEN_EXISTING).
+
+%%--------------------------------------------------------------------
+open_file_dir_v5(Config) when is_list(Config) ->
+    PrivDir =  proplists:get_value(priv_dir, Config),
+    FileName = "open_file_dir_v5",
+    FilePath = filename:join(PrivDir, FileName),
+    ok = filelib:ensure_dir(FilePath),
+    ok = file:make_dir(FilePath),
+    {Cm, Channel} = proplists:get_value(sftp, Config),
+    ReqId = 0,
+    {ok, <<?SSH_FXP_STATUS, ?UINT32(ReqId),
+	   ?UINT32(?SSH_FX_FAILURE), _/binary>>, _} =
+        open_file(FileName, Cm, Channel, ReqId,
+                  ?ACE4_READ_DATA  bor ?ACE4_READ_ATTRIBUTES,
+                  ?SSH_FXF_OPEN_EXISTING).
+
+%%--------------------------------------------------------------------
+open_file_dir_v6(Config) when is_list(Config) ->
+    PrivDir =  proplists:get_value(priv_dir, Config),
+    FileName = "open_file_dir_v6",
+    FilePath = filename:join(PrivDir, FileName),
+    ok = filelib:ensure_dir(FilePath),
+    ok = file:make_dir(FilePath),
+    {Cm, Channel} = proplists:get_value(sftp, Config),
+    ReqId = 0,
+    {ok, <<?SSH_FXP_STATUS, ?UINT32(ReqId),
+	   ?UINT32(?SSH_FX_FILE_IS_A_DIRECTORY), _/binary>>, _} =
+        open_file(FileName, Cm, Channel, ReqId,
+                  ?ACE4_READ_DATA  bor ?ACE4_READ_ATTRIBUTES,
+                  ?SSH_FXF_OPEN_EXISTING).
+
 %%--------------------------------------------------------------------
 %% Internal functions ------------------------------------------------
 %%--------------------------------------------------------------------
 prep(Config) ->
-    PrivDir =  ?config(priv_dir, Config),
+    PrivDir =  proplists:get_value(priv_dir, Config),
     TestFile = filename:join(PrivDir, "test.txt"),
     TestFile1 = filename:join(PrivDir, "test1.txt"),
 
@@ -655,7 +806,7 @@ prep(Config) ->
     file:delete(TestFile1),
 
     %% Initial config
-    DataDir = ?config(data_dir, Config),
+    DataDir = proplists:get_value(data_dir, Config),
     FileName = filename:join(DataDir, "test.txt"),
     file:copy(FileName, TestFile),
     Mode = 8#00400 bor 8#00200 bor 8#00040, % read & write owner, read group
@@ -681,11 +832,11 @@ reply(Cm, Channel, RBuf) ->
 	    closed;
 	{ssh_cm, Cm, Msg} ->
 	    ct:fail(Msg)
+    after 
+	90000 -> ct:fail("timeout ~p:~p",[?MODULE,?LINE])
     end.
 
-
 open_file(File, Cm, Channel, ReqId, Access, Flags) ->
-
     Data = list_to_binary([?uint32(ReqId),
 			   ?binary(list_to_binary(File)),
 			   ?uint32(Access),
@@ -784,7 +935,7 @@ read_dir(Handle, Cm, Channel, ReqId) ->
     case reply(Cm, Channel) of
 	{ok, <<?SSH_FXP_NAME, ?UINT32(ReqId), ?UINT32(Count),
 	       ?UINT32(Len), Listing:Len/binary, _/binary>>, _} ->
-	    ct:pal("Count: ~p Listing: ~p~n",
+	    ct:log("Count: ~p Listing: ~p~n",
 			       [Count, binary_to_list(Listing)]),
 	    read_dir(Handle, Cm, Channel, ReqId);
 	{ok, <<?SSH_FXP_STATUS, ?UINT32(ReqId),
