@@ -1,7 +1,9 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2007-2023. All Rights Reserved.
+%% SPDX-License-Identifier: Apache-2.0
+%%
+%% Copyright Ericsson AB 2007-2025. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -24,6 +26,7 @@
 %%----------------------------------------------------------------------
 
 -module(tls_handshake_1_3).
+-moduledoc false.
 
 -include("tls_handshake_1_3.hrl").
 -include("ssl_alert.hrl").
@@ -58,13 +61,13 @@
 -export([process_certificate_request/2,
          process_certificate/2,
          calculate_handshake_secrets/5,
+         calculate_exporter_master_secret/1,
          verify_certificate_verify/2,
          validate_finished/2,
-         maybe_calculate_resumption_master_secret/1,
          replace_ch1_with_message_hash/1,
          select_common_groups/2,
          verify_signature_algorithm/2,
-         forget_master_secret/1,
+         handle_secrets/1,
          set_client_random/2,
          handle_pre_shared_key/3,
          update_start_state/2,
@@ -82,6 +85,8 @@
          calculate_traffic_secrets/1,
          calculate_client_early_traffic_secret/5,
          calculate_client_early_traffic_secret/2,
+         early_data_secret/1,
+         hs_traffic_secrets/1,
          encode_early_data/2,
          get_ticket_data/3,
          ciphers_for_early_data/1,
@@ -108,7 +113,6 @@ server_hello(MsgType, SessionId, KeyShare, PSK, ConnectionStates) ->
     Extensions = server_hello_extensions(MsgType, KeyShare, PSK),
     #server_hello{server_version = ?LEGACY_VERSION, %% legacy_version
 		  cipher_suite = SecParams#security_parameters.cipher_suite,
-                  compression_method = 0, %% legacy attribute
 		  random = server_hello_random(MsgType, SecParams),
 		  session_id = SessionId,
 		  extensions = Extensions
@@ -244,7 +248,9 @@ add_signature_algorithms_cert(Extensions, SignAlgsCert) ->
 
 filter_tls13_algs(undefined) -> undefined;
 filter_tls13_algs(Algo) ->
-    lists:foldl(fun(Atom, Acc) when is_atom(Atom) ->
+    lists:foldl(fun(default, Acc) ->
+                        Acc;
+                   (Atom, Acc) when is_atom(Atom) ->
                         [Atom | Acc];
                    ({sha512, rsa}, Acc) ->
                         [rsa_pkcs1_sha512 | Acc];
@@ -384,13 +390,11 @@ create_change_cipher_spec(#state{ssl_options = #{log_level := LogLevel}}) ->
     %% Dummy connection_states with NULL cipher
     ConnectionStates =
         #{current_write =>
-              #{compression_state => undefined,
-                cipher_state => undefined,
+              #{cipher_state => undefined,
                 sequence_number => 1,
                 security_parameters =>
                     #security_parameters{
                        bulk_cipher_algorithm = 0,
-                       compression_algorithm = ?NULL,
                        mac_algorithm = ?NULL
                       },
                 mac_secret => undefined}},
@@ -409,6 +413,7 @@ process_certificate_request(#certificate_request_1_3{
                                    connection_env = #connection_env{cert_key_alts = CertKeyAlts,
                                                                     negotiated_version = Version},
                                    static_env = #static_env{cert_db = CertDbHandle, cert_db_ref = CertDbRef},
+                                   handshake_env = HsEnv,
                                    session = Session0} =
                                 State) ->
     ServerSignAlgs = get_signature_scheme_list(
@@ -421,7 +426,8 @@ process_certificate_request(#certificate_request_1_3{
     Session = select_client_cert_key_pair(Session0, CertKeyPairs,
                                           ServerSignAlgs, ServerSignAlgsCert, ClientSignAlgs,
                                           CertDbHandle, CertDbRef, CertAuths, undefined),
-    {ok, {State#state{client_certificate_status = requested, session = Session}, wait_cert}}.
+    {ok, {State#state{handshake_env = HsEnv#handshake_env{client_certificate_status = requested},
+                      session = Session}, wait_cert}}.
 
 process_certificate(#certificate_1_3{
                        certificate_request_context = <<>>,
@@ -452,9 +458,10 @@ process_certificate(#certificate_1_3{certificate_list = CertEntries},
                                   crl_db = CRLDbHandle},
                            handshake_env =
                                #handshake_env{
-                                  ocsp_stapling_state = OcspState}} = State0) ->
-    case validate_certificate_chain(CertEntries, CertDbHandle, CertDbRef,
-                                    SslOptions, CRLDbHandle, Role, Host, OcspState) of
+                                  stapling_state = StaplingState}} = State0) ->
+    case validate_certificate_chain(
+           CertEntries, CertDbHandle, CertDbRef, SslOptions, CRLDbHandle, Role,
+           Host, StaplingState) of
         #alert{} = Alert ->
             State = update_encryption_state(Role, State0),
             {error, {Alert, State}};
@@ -576,14 +583,14 @@ encode_handshake(HandshakeMsg) ->
 
 encode_early_data(Cipher,
                   #state{
-                     flight_buffer = Flight0,
+                     handshake_env = #handshake_env{
+                                        flight_buffer = Flight0} = HsEnv,
                      protocol_specific = #{sender := _Sender},
+                     connection_states = ConnectionStates0,
                      ssl_options = #{versions := [Version|_],
                                      early_data := EarlyData} = _SslOpts0
                     } = State0) ->
-    #state{connection_states =
-               #{current_write :=
-                     #{security_parameters := SecurityParameters0} = Write0} = ConnectionStates0} = State0,
+    #{current_write := #{security_parameters := SecurityParameters0} = Write0} = ConnectionStates0,
     BulkCipherAlgo = ssl_cipher:bulk_cipher_algorithm(Cipher),
     SecurityParameters = SecurityParameters0#security_parameters{
                            cipher_type = ?AEAD,
@@ -592,7 +599,7 @@ encode_early_data(Cipher,
     ConnectionStates1 = ConnectionStates0#{current_write => Write},
     {BinEarlyData, ConnectionStates} = tls_record:encode_data([EarlyData], Version, ConnectionStates1),
     State0#state{connection_states = ConnectionStates,
-		 flight_buffer = Flight0 ++ [BinEarlyData]}.
+                 handshake_env = HsEnv#handshake_env{flight_buffer = Flight0 ++ [BinEarlyData]}}.
 
 
 %%====================================================================
@@ -602,7 +609,7 @@ encode_early_data(Cipher,
 
 decode_handshake(?SERVER_HELLO, <<?BYTE(Major), ?BYTE(Minor), Random:32/binary,
                                   ?BYTE(SID_length), Session_ID:SID_length/binary,
-                                  Cipher_suite:2/binary, ?BYTE(Comp_method),
+                                  Cipher_suite:2/binary, ?BYTE(_CompMethod),
                                   ?UINT16(ExtLen), Extensions:ExtLen/binary>>)
   when Random =:= ?HELLO_RETRY_REQUEST_RANDOM ->
     HelloExtensions = ssl_handshake:decode_hello_extensions(Extensions, ?TLS_1_3, {Major, Minor},
@@ -612,7 +619,6 @@ decode_handshake(?SERVER_HELLO, <<?BYTE(Major), ?BYTE(Minor), Random:32/binary,
        random = Random,
        session_id = Session_ID,
        cipher_suite = Cipher_suite,
-       compression_method = Comp_method,
        extensions = HelloExtensions};
 decode_handshake(?CERTIFICATE_REQUEST, <<?BYTE(0), ?UINT16(Size), EncExts:Size/binary>>) ->
     Exts = decode_extensions(EncExts, certificate_request),
@@ -808,42 +814,44 @@ update_encryption_state(client, State) ->
 
 
 validate_certificate_chain(CertEntries, CertDbHandle, CertDbRef,
-                           SslOptions, CRLDbHandle, Role, Host, OcspState0) ->
-    {Certs, CertExt, OcspState} = split_cert_entries(CertEntries, OcspState0),
-    OcspResponderCerts =
-        case maps:get(ocsp_stapling, SslOptions, disabled) of
-            #{ocsp_responder_certs := V} ->
-                V;
-            disabled ->
-                ?DEFAULT_OCSP_RESPONDER_CERTS
-        end,
-    ssl_handshake:certify(#certificate{asn1_certificates = Certs}, CertDbHandle, CertDbRef,
-                          SslOptions, CRLDbHandle, Role, Host, ?TLS_1_3,
-                          #{cert_ext => CertExt,
-                            ocsp_state => OcspState,
-                            ocsp_responder_certs => OcspResponderCerts}).
+                           SslOptions, CRLDbHandle, Role, Host, StaplingState) ->
+    try split_cert_entries(CertEntries, StaplingState, [], #{}) of
+        {Certs, ExtInfo} ->
+            ssl_handshake:certify(Certs, CertDbHandle,
+                                  CertDbRef, SslOptions, CRLDbHandle, Role, Host, ?TLS_1_3,
+                                  ExtInfo)
+    catch error:{_,{error, {asn1, Asn1Reason}}}=Reason:ST ->
+            %% ASN-1 decode of certificate somehow failed
+            ?SSL_LOG(info, asn1_decode, [Reason, {stacktrace, ST}]),
+            ?ALERT_REC(?FATAL, ?CERTIFICATE_UNKNOWN, {failed_to_decode_certificate, Asn1Reason})
+    end.
 
-store_peer_cert(#state{session = Session,
-                       handshake_env = HsEnv} = State, PeerCert, PublicKeyInfo) ->
-    State#state{session = Session#session{peer_certificate = PeerCert},
+store_peer_cert(#state{session = Session, handshake_env = HsEnv} = State,
+                PeerCert, PublicKeyInfo) ->
+    State#state{session = Session#session{peer_certificate = PeerCert#cert.der},
                 handshake_env = HsEnv#handshake_env{public_key_info = PublicKeyInfo}}.
 
-split_cert_entries(CertEntries, OcspState) ->
-    split_cert_entries(CertEntries, OcspState, [], #{}).
+split_cert_entries([], StaplingState, Chain, CertExt) ->
+    {lists:reverse(Chain), #{cert_ext => CertExt,
+                             stapling_state => StaplingState}};
+split_cert_entries([#certificate_entry{data = DerCert,
+                                       extensions = Extensions0} | CertEntries],
+                   #{configured := StaplingConfigured} = StaplingState0, Chain,
+                   CertExt) ->
+    OtpCert = public_key:pkix_decode_cert(DerCert, otp),
+    Cert = #cert{der=DerCert, otp=OtpCert},
 
-split_cert_entries([], OcspState, Chain, Ext) ->
-    {lists:reverse(Chain), Ext, OcspState};
-split_cert_entries([#certificate_entry{data = DerCert, extensions = Extensions0} | CertEntries],
-                   OcspState0, Chain, Ext) ->
     Id = public_key:pkix_subject_id(DerCert),
     Extensions = [ExtValue || {_, ExtValue} <- maps:to_list(Extensions0)],
-    OcspState = case maps:get(status_request, Extensions0, undefined) of
-                    undefined ->
-                        OcspState0;
-                    _ ->
-                        OcspState0#{ocsp_expect => stapled}
-                end,
-    split_cert_entries(CertEntries, OcspState, [DerCert | Chain], Ext#{Id => Extensions}).
+    StaplingState = case {maps:get(status_request, Extensions0, undefined),
+                          StaplingConfigured} of
+                        {undefined, _} ->
+                            StaplingState0;
+                        {_, true} ->
+                            StaplingState0#{status => received_staple}
+                    end,
+    split_cert_entries(CertEntries, StaplingState, [Cert | Chain],
+                       CertExt#{Id => Extensions}).
 
 %% 4.4.1.  The Transcript Hash
 %%
@@ -885,6 +893,12 @@ message_hash(ClientHello1, HKDFAlgo) ->
      0,0,ssl_cipher:hash_size(HKDFAlgo),
      crypto:hash(HKDFAlgo, ClientHello1)].
 
+handle_secrets(State0) ->
+    State1 = calculate_traffic_secrets(State0),
+    State2 = #state{protocol_specific = PS} = maybe_calculate_resumption_master_secret(State1),
+    ExporterSecret = calculate_exporter_master_secret(State2),
+    State3 =  State2#state{protocol_specific = PS#{exporter_master_secret => ExporterSecret}},
+    forget_master_secret(State3).
 
 calculate_handshake_secrets(PublicKey, PrivateKey, SelectedGroup, PSK,
                               #state{connection_states = ConnectionStates,
@@ -916,7 +930,7 @@ calculate_handshake_secrets(PublicKey, PrivateKey, SelectedGroup, PSK,
     ReadFinishedKey = tls_v1:finished_key(ClientHSTrafficSecret, HKDFAlgo),
     WriteFinishedKey = tls_v1:finished_key(ServerHSTrafficSecret, HKDFAlgo),
 
-    State1 = maybe_store_handshake_traffic_secret(State0, ClientHSTrafficSecret, ServerHSTrafficSecret),
+    State1 = maybe_store_handshake_traffic_secret(State0, HKDFAlgo, ClientHSTrafficSecret, ServerHSTrafficSecret),
 
     update_pending_connection_states(State1, HandshakeSecret, undefined,
                                      undefined, undefined,
@@ -954,7 +968,7 @@ calculate_client_early_traffic_secret(
         client ->
             PendingWrite0 = ssl_record:pending_connection_state(ConnectionStates, write),
             PendingWrite1 = maybe_store_early_data_secret(Opts, ClientEarlyTrafficSecret,
-                                                          PendingWrite0),
+                                                          PendingWrite0, HKDFAlgo),
             PendingWrite = update_connection_state(PendingWrite1, undefined, undefined,
                                                    undefined,
                                                    Key, IV, undefined),
@@ -962,21 +976,40 @@ calculate_client_early_traffic_secret(
         server ->
             PendingRead0 = ssl_record:pending_connection_state(ConnectionStates, read),
             PendingRead1 = maybe_store_early_data_secret(Opts, ClientEarlyTrafficSecret,
-                                                         PendingRead0),
+                                                         PendingRead0, HKDFAlgo),
             PendingRead = update_connection_state(PendingRead1, undefined, undefined,
                                                    undefined,
                                                    Key, IV, undefined),
             State0#state{connection_states = ConnectionStates#{pending_read => PendingRead}}
     end.
 
+early_data_secret(undefined) ->
+    [];
+early_data_secret(Secret) ->
+    [{client_early_data_secret, Secret}].
 
+hs_traffic_secrets(#{client_handshake_traffic_secret := ClientHSTrafficSecret,
+                      server_handshake_traffic_secret := ServerHSTrafficSecret}) ->
+     [{client_handshake_traffic_secret, ClientHSTrafficSecret},
+      {server_handshake_traffic_secret, ServerHSTrafficSecret}];
+hs_traffic_secrets(_) ->
+    [].
 
-maybe_store_early_data_secret(#{keep_secrets := true}, EarlySecret, State) ->
-    #{security_parameters := SecParams0} = State,
-    SecParams = SecParams0#security_parameters{client_early_data_secret = EarlySecret},
-    State#{security_parameters := SecParams};
-maybe_store_early_data_secret(_, _, State) ->
-    State.
+maybe_store_early_data_secret(#{keep_secrets := Keep}, EarlySecret,
+                              #{security_parameters := SecParams0} = CSState,
+                              Prf) when Keep =/= false ->
+    case Keep of
+        {keylog, Fun} ->
+            ClientRand = SecParams0#security_parameters.client_random,
+            KeyLog = ssl_logger:keylog_early_data(ClientRand, Prf, EarlySecret),
+            ssl_logger:keylog(KeyLog, ClientRand, Fun),
+            CSState;
+        _ ->
+            SecParams = SecParams0#security_parameters{client_early_data_secret = EarlySecret},
+            CSState#{security_parameters := SecParams}
+    end;
+maybe_store_early_data_secret(_, _, CSState, _) ->
+    CSState.
 
 %% Server
 %% get_pre_shared_key(undefined, HKDFAlgo) ->
@@ -1089,9 +1122,6 @@ calculate_traffic_secrets(#state{
                                      ReadKey, ReadIV, undefined,
                                      WriteKey, WriteIV, undefined).
 
-
-
-
 %% X25519, X448
 calculate_shared_secret(OthersKey, MyKey, Group)
   when is_binary(OthersKey) andalso is_binary(MyKey) andalso
@@ -1117,8 +1147,8 @@ maybe_calculate_resumption_master_secret(#state{
                              ssl_options = #{session_tickets := SessionTickets},
                              connection_states = ConnectionStates,
                              handshake_env =
-                                 #handshake_env{
-                                    tls_handshake_history = HHistory}} = State)
+                                                #handshake_env{
+                                                   tls_handshake_history = HHistory}} = State)
   when SessionTickets =/= disabled  ->
     #{security_parameters := SecParamsR} =
         ssl_record:pending_connection_state(ConnectionStates, read),
@@ -1128,6 +1158,18 @@ maybe_calculate_resumption_master_secret(#state{
     RMS = tls_v1:resumption_master_secret(HKDFAlgo, MasterSecret, lists:reverse(Messages0)),
     update_resumption_master_secret(State, RMS).
 
+calculate_exporter_master_secret(#state{
+                             static_env = #static_env{role = Role},
+                             connection_states = ConnectionStates,
+                             handshake_env =
+                                 #handshake_env{
+                                    tls_handshake_history = HHistory}}) ->
+    #{security_parameters := SecParamsR} =
+        ssl_record:pending_connection_state(ConnectionStates, read),
+    #security_parameters{prf_algorithm = HKDFAlgo,
+                         master_secret = MasterSecret} = SecParamsR,
+    Messages = get_handshake_context(Role, HHistory),
+    tls_v1:exporter_master_secret(HKDFAlgo, MasterSecret, lists:reverse(Messages)).
 
 forget_master_secret(#state{connection_states =
                                 #{pending_read := PendingRead,
@@ -1162,12 +1204,21 @@ overwrite_client_random(ConnectionState = #{security_parameters := SecurityParam
 
 
 maybe_store_handshake_traffic_secret(#state{connection_states =
-                                                #{pending_read := PendingRead} = CS,
-                                            ssl_options = #{keep_secrets := true}} = State,
-                                     ClientHSTrafficSecret, ServerHSTrafficSecret) ->
-    PendingRead1 = store_handshake_traffic_secret(PendingRead, ClientHSTrafficSecret, ServerHSTrafficSecret),
-    State#state{connection_states = CS#{pending_read => PendingRead1}};
-maybe_store_handshake_traffic_secret(State, _, _) ->
+                                                #{pending_read := PendingRead0}= CS,
+                                            ssl_options = #{keep_secrets := Keep}} = State, Prf,
+                                     ClientHSTrafficSecret, ServerHSTrafficSecret) when Keep =/= false ->
+     case Keep of
+        {keylog, Fun} ->
+             #{security_parameters := SecParams0} = PendingRead0,
+             ClientRand = SecParams0#security_parameters.client_random,
+             KeyLog = ssl_logger:keylog_hs(ClientRand, Prf, ClientHSTrafficSecret, ServerHSTrafficSecret),
+             ssl_logger:keylog(KeyLog, ClientRand, Fun),
+             State;
+         _ ->
+             PendingRead = store_handshake_traffic_secret(PendingRead0, ClientHSTrafficSecret, ServerHSTrafficSecret),
+             State#state{connection_states = CS#{pending_read => PendingRead}}
+     end;
+maybe_store_handshake_traffic_secret(State, _, _, _) ->
     State.
 
 store_handshake_traffic_secret(ConnectionState, ClientHSTrafficSecret, ServerHSTrafficSecret) ->
@@ -1232,28 +1283,34 @@ update_start_state(State, Map) ->
     SelectedSignAlg = maps:get(sign_alg, Map, undefined),
     PeerPublicKey = maps:get(peer_public_key, Map, undefined),
     ALPNProtocol = maps:get(alpn, Map, undefined),
+    Random = maps:get(random, Map),
     update_start_state(State, Cipher, KeyShare, SessionId,
                        Group, SelectedSignAlg, PeerPublicKey,
-                       ALPNProtocol).
+                       ALPNProtocol, Random).
 %%
 update_start_state(#state{connection_states = ConnectionStates0,
                           handshake_env = #handshake_env{} = HsEnv,
+                          static_env = #static_env{role = Role},
                           connection_env = CEnv,
                           session = Session} = State,
                    Cipher, KeyShare, SessionId,
-                   Group, SelectedSignAlg, PeerPublicKey, ALPNProtocol) ->
+                   Group, SelectedSignAlg, PeerPublicKey, ALPNProtocol, Random) ->
     #{security_parameters := SecParamsR0} = PendingRead =
         maps:get(pending_read, ConnectionStates0),
     #{security_parameters := SecParamsW0} = PendingWrite =
         maps:get(pending_write, ConnectionStates0),
-    SecParamsR = ssl_cipher:security_parameters_1_3(SecParamsR0, Cipher),
-    SecParamsW = ssl_cipher:security_parameters_1_3(SecParamsW0, Cipher),
+    SecParamsR1 = ssl_cipher:security_parameters_1_3(SecParamsR0, Cipher),
+    SecParamsW1 = ssl_cipher:security_parameters_1_3(SecParamsW0, Cipher),
+
+    SecParamsR = update_random(Role, SecParamsR1, Random),
+    SecParamsW = update_random(Role, SecParamsW1, Random),
+
     ConnectionStates =
         ConnectionStates0#{pending_read => PendingRead#{security_parameters => SecParamsR},
                            pending_write => PendingWrite#{security_parameters => SecParamsW}},
     State#state{connection_states = ConnectionStates,
-                handshake_env = HsEnv#handshake_env{alpn = ALPNProtocol},
-                key_share = KeyShare,
+                handshake_env = HsEnv#handshake_env{alpn = ALPNProtocol,
+                                                    key_share = KeyShare},
                 session = Session#session{session_id = SessionId,
                                           ecc = Group,
                                           sign_alg = SelectedSignAlg,
@@ -1261,6 +1318,10 @@ update_start_state(#state{connection_states = ConnectionStates0,
                                           cipher_suite = Cipher},
                 connection_env = CEnv#connection_env{negotiated_version = ?TLS_1_3}}.
 
+update_random(server, SParams, Random) ->
+    SParams#security_parameters{client_random = Random};
+update_random(client, SParams, Random) ->
+    SParams#security_parameters{server_random = Random}.
 
 update_resumption_master_secret(#state{connection_states = ConnectionStates0} = State,
                                 ResumptionMasterSecret) ->
@@ -1863,11 +1924,11 @@ path_validation(TrustedCert, Path, ServerName, Role, CertDbHandle, CertDbRef, CR
                   customize_hostname_check := CustomizeHostnameCheck,
                   crl_check := CrlCheck,
                   log_level := LogLevel,
+                  allow_any_ca_purpose := AllowAnyPurpose,
                   signature_algs := SignAlgos,
                   signature_algs_cert := SignAlgosCert} = Opts,
                 #{cert_ext := CertExt,
-                  ocsp_responder_certs := OcspResponderCerts,
-                  ocsp_state := OcspState}) ->
+                  stapling_state := StaplingState}) ->
     ValidationFunAndState =
         ssl_handshake:validation_fun_and_state(VerifyFun,
                                                #{role => Role,
@@ -1884,12 +1945,12 @@ path_validation(TrustedCert, Path, ServerName, Role, CertDbHandle, CertDbRef, CR
                                                  version => Version,
                                                  issuer => TrustedCert,
                                                  cert_ext => CertExt,
-                                                 ocsp_responder_certs => OcspResponderCerts,
-                                                 ocsp_state => OcspState
+                                                 stapling_state => StaplingState,
+                                                 allow_any_ca_purpose => AllowAnyPurpose,
+                                                 path_len => length(Path)
                                                 },
                                                Path, LogLevel),
-    Options = [{max_path_length, maps:get(depth, Opts, ?DEFAULT_DEPTH)},
-               {verify_fun, ValidationFunAndState}],
+    Options = ssl_handshake:path_validation_options(Opts, ValidationFunAndState),
     public_key:pkix_path_validation(TrustedCert, Path, Options).
 
 select_client_cert_key_pair(Session0,
